/* spfmilter - SPF mail filter module
**
** Copyright  2004 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** Tweaked to work with later versions of libspf2 by Ged Haywood, November 2008.
** No rights reserved. :)  If you wonder where my email address is, ask yourself
** why you're looking at this code in the first place.  There is a mailing list.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**	notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**	notice, this list of conditions and the following disclaimer in the
**	documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
**
** For commentary on this license please see http://www.acme.com/license.html
*/

#if ! ( defined(USE_LIBSPF) || defined(USE_LIBSPF2) )
#define USE_LIBSPF2	/* default library is libspf2 */
#endif

#if defined(USE_LIBSPF) && defined(USE_LIBSPF2)
#error "both USE_LIBSPF and USE_LIBSPF2 are defined - please pick one"
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_GRP_H
#include <grp.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include <libmilter/mfapi.h>

#include "version.h"
#include "match.h"
#include "iparray.h"


/* Defines. */

#define HEADER_NAME "Received-SPF"

#define SPFMILTER_RESULT_PASS 0
#define SPFMILTER_RESULT_FAIL 1
#define SPFMILTER_RESULT_SOFTFAIL 2
#define SPFMILTER_RESULT_NEUTRAL 3
#define SPFMILTER_RESULT_UNKNOWN 4
#define SPFMILTER_RESULT_UNMECH 5
#define SPFMILTER_RESULT_ERROR 6
#define SPFMILTER_RESULT_NONE 7
#define SPFMILTER_RESULT_INVALID 8

#define SPFMILTER_ACTION_UNKNOWN 0
#define SPFMILTER_ACTION_REJECT 1
#define SPFMILTER_ACTION_MARK 2
#define SPFMILTER_ACTION_TEMPFAIL 3


/* Structs. */


/* Fallback list. */
struct lib_fallback_s;
typedef struct lib_fallback_s lib_fallback_t;
typedef struct {
    const char* pattern;
    const char* str;
    lib_fallback_t* lib_fallback;
    } fallback_t;
static fallback_t* fallbacks;
static int max_fallbacks, num_fallbacks;


/* Per-connection data structure. */
#define DEL_HEADER_MAX 10
struct lib_data_s;
typedef struct lib_data_s lib_data_t;
typedef struct {
    int whitelisted;
    int authenticated;
    const char* ip_str;
    const char* helo;
    const char* mail_from;
    const char* use_mail_from;
    lib_data_t* lib_data;
    int result;
    int action;
    int n_spf_headers;
    int n_del_headers;
    int del_header_list[DEL_HEADER_MAX];
    char buf1[1000];
    char buf2[2000];
    } connection_data_t;

/* A word about buf1 and buf2, above.  These are scratch buffers for
** temporary use by any routine that needs them.  Normally we'd just
** declare them locally, on the stack.  The problem is that we are
** running in a thread, each thread has a separate stack of fixed size,
** we don't know what that size is, and if we exceed it we dump core.
** So we really want to avoid using up the stack with large local variable
** frames; instead, we use the per-connection data object which is
** allocated from the heap.  It makes the connection objects fairly
** large, but not unreasonably so.
**
** We do have to be a little careful using the buffers, so that two
** routines in the same call stack don't try to make use of them
** simultaneously.  Indeed, there's one case where we need two buffers
** at the same time so that's why we have two here.
*/


/* Forwards. */

static void usage( void );
static void init_fallback( const char* fallback_filename, const char* guess_str );
static void add_fallback( const char* pattern, const char* str );
static void fini_fallback( void );
static int find_fallback( const char* from );
static void init_whitelist( const char* whitelist_filename );
static void fini_whitelist( void );
static void trim( char* str );
static void init_uid( const char* user );
static void init_socket( const char* sockpath );
static void fini_socket( const char* sockpath );
static void unlink_socket( const char* sockpath );
static void init_daemon( void );
static void init_pidfile( const char* pidfile );
static void fini_pidfile( const char* pidfile );

/* The milter calback routines.  Signatures for these are fixed
** by the milter API.
*/
static sfsistat spf_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr );
static sfsistat spf_helo( SMFICTX* ctx, char* helohost );
static sfsistat spf_envfrom( SMFICTX* ctx, char** fromargs );
static sfsistat spf_envrcpt( SMFICTX* ctx, char** rcptargs );
static sfsistat spf_header( SMFICTX* ctx, char* name, char* value );
static sfsistat spf_eoh( SMFICTX* ctx );
static sfsistat spf_eom( SMFICTX* ctx );
static sfsistat spf_abort( SMFICTX* ctx );
static sfsistat spf_close( SMFICTX* ctx );

static char* trim_address( char* address );
static sfsistat handle_result( SMFICTX* ctx );
static void escape_percents( const char* in, char* out, int out_size );
static int receiver_is_me( const char* value );
static void build_header( connection_data_t* cd, char* header, size_t header_size, const char* reason );
static connection_data_t* init_connection_data( void );
static int init_message_data( connection_data_t* cd );
static void fini_message_data( connection_data_t* cd );
static void fini_connection_data( connection_data_t* cd );
static char* result_str( int result );
static void reason_str( connection_data_t* cd, char* str, size_t size );

/* Library interface routines.  These are implemented in parallel for
** both libspf and libspf2.
*/
static int lib_init( void );
static const char* lib_version( void );
static lib_fallback_t* lib_init_fallback( const char* str );
static int lib_set_local_hostname(		 lib_data_t* ld );
static lib_data_t* lib_init_connection_data( void );
static int lib_init_message_data(		 lib_data_t* ld );
static int lib_set_ipv4(			 lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str );
static int lib_set_ipv6(			 lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str );
static int lib_set_helo_hostname(		 lib_data_t* ld, char* helo_hostname );
static int lib_set_from(			 lib_data_t* ld, const char* from );
static int lib_do_check_m(			 lib_data_t* ld, const char* from );
static int lib_do_check_recipient(		 lib_data_t* ld, const char* to );
static int lib_do_check_final(			 lib_data_t* ld );
static int lib_get_result(			 lib_data_t* ld );

static const char* lib_get_explanation(		 lib_data_t* ld );
static const char* lib_get_error(		 lib_data_t* ld );
static void lib_fini_message_data(		 lib_data_t* ld );
static void lib_fini_connection_data(		 lib_data_t* ld );
static void lib_fini_fallback( lib_fallback_t* lf );
static void lib_fini( void );


/* Globals. */

static char* argv0;

static char* localpolicy_str;
static int trustedforwarders;
static iparray whitelist;
int recipientmx;
static char* explanation_str;
static int markonly;
static int debug;
static char* local_hostname;
static int local_hostname_len;

static int header_name_len;

static struct smfiDesc smfilter = {
    "SPF",				/* filter name */
    SMFI_VERSION,			/* version code -- do not change */
    SMFIF_CHGHDRS|SMFIF_ADDHDRS,	/* flags */
    spf_connect,			/* connection info filter */
    spf_helo,				/* SMTP HELO command filter */
    spf_envfrom,			/* envelope sender filter */
    spf_envrcpt,			/* envelope recipient filter */
    spf_header,				/* header filter */
    spf_eoh,				/* end of header */
    NULL,				/* body block filter */
    spf_eom,				/* end of message */
    spf_abort,				/* message aborted */
    spf_close				/* connection cleanup */
    };

#ifdef HAVE_GETOPT_LONG
static const struct option	longopts[] = {
    { "localpolicy",		required_argument,	NULL,	'l', },
    { "trustedforwarders",	no_argument,		NULL,	't', },
    { "guess",			required_argument,	NULL,	'g', },
    { "fallback",		required_argument,	NULL,	'f', },
    { "whitelist",		required_argument,	NULL,	'w', },
    { "recipientmx",		no_argument,		NULL,	'r', },
    { "explanation",		required_argument,	NULL,	'e', },
    { "markonly",		no_argument,		NULL,	'm', },
    { "user",			required_argument,	NULL,	'u', },
    { "pidfile",		required_argument,	NULL,	'p', },
    { "nodaemon",		no_argument,		NULL,	'X', },
    { "help",			no_argument,		NULL,	'h', },
    { "debug",			optional_argument,	NULL,	'd', },
    { 0, 0, 0, 0 },
};
#define DOC_LONGOPT(l, v, t, p1) \
    do { \
	(void) fprintf( stderr, "    --%s%c%s%*s" t "\n", l, (v ? '=' : ' '),  (v ? v : ""), p1, "" ); \
    } while( 0 )
#else /* HAVE_GETOPT_LONG */
#define DOC_LONGOPT(l, v, t, p1) do { } while( 0 )
#endif /* HAVE_GETOPT_LONG */
static const char* shortopts = "l:tg:f:w:re:mu:p:Xhd::";

#define DOC_OPT(s, l, v, t, p0, p1) \
    do { \
	(void) fprintf( stderr, "    -%c%c%s%*s" t "\n", s, (v ? ' ' : ' '), (v ? v : ""), p0, "" ); \
	DOC_LONGOPT(l, v, t, p1); \
    } while( 0 )


int
main( int argc, char** argv )
{
    char* guess_str;
    char* fallback_filename;
    char* whitelist_filename;
    char* user;
    char* pidfile;
    int nodaemon;
    char* sockpath;
#ifdef HAVE_GETOPT_LONG
    int idx;
#endif /* HAVE_GETOPT_LONG */
    char c;

    localpolicy_str = (char*) 0;
    trustedforwarders = 0;
    guess_str = (char*) 0;
    fallback_filename = (char*) 0;
    whitelist_filename = (char*) 0;
    recipientmx = 0;
    explanation_str = (char*) 0;
    markonly = 0;
    user = (char*) 0;
    pidfile = (char*) 0;
    nodaemon = 0;
    debug = 0;

    /* Figure out the program's name. */
    argv0 = strrchr( argv[0], '/' );
    if( argv0 != (char*) 0 )
	++argv0;
    else
	argv0 = argv[0];

    openlog( argv0, LOG_PID, LOG_MAIL );

    syslog( LOG_DEBUG, "main() - spfmilter called as [%s]", argv0 );

    header_name_len = strlen( HEADER_NAME );

    /* Parse args. */
    while ( ( c =
#ifdef HAVE_GETOPT_LONG
		  getopt_long( argc, argv, shortopts, longopts, &idx )
#else /* HAVE_GETOPT_LONG */
		  getopt( argc, argv, shortopts )
#endif /* HAVE_GETOPT_LONG */
			) != -1 )
	{
	switch (c)
	    {
	    case 'l':
		localpolicy_str = optarg;
		break;
	    case 't':
		trustedforwarders = 1;
		break;
	    case 'g':
		guess_str = optarg;
		break;
	    case 'f':
		fallback_filename = optarg;
		break;
	    case 'w':
		whitelist_filename = optarg;
		break;
	    case 'r':
		recipientmx = 1;
		break;
	    case 'e':
		explanation_str = optarg;
		break;
	    case 'm':
		markonly = 1;
		break;
	    case 'u':
		user = optarg;
		break;
	    case 'p':
		pidfile = optarg;
		break;
	    case 'X':
		nodaemon = 1;
		break;
	    case 'h':
	    case '?':
		usage();
		exit( 1 );
	    case 'd':
		if( optarg )
		    debug = atoi( optarg );
		else
		    debug = 1;
		break;
	    default:
		(void) fprintf( stderr, "Unrecognised option '%c'\n", c );
		exit( 1 );
	    }
	}
    argv += optind;
    argc -= optind;

    syslog( LOG_DEBUG, "main() - argc = %d", argc );

    if( argc != 1 )
	{
	usage();
	exit( 1 );
	}
    sockpath = argv[0];

    syslog( LOG_DEBUG, "main() - parsed arguments" );

    local_hostname = (char*) 0;
    if( ! lib_init() )
	{
	(void) fprintf( stderr, "%s: lib_init() failed\n", argv0 );
	exit( 1 );
	}


    syslog( LOG_DEBUG, "main() - lib_init() completed" );

    init_fallback( fallback_filename, guess_str );

    syslog( LOG_DEBUG, "main() - init_fallback() completed" );

    init_whitelist( whitelist_filename );

    syslog( LOG_DEBUG, "main() - init_whitelist() completed" );

    init_uid( user );

    syslog( LOG_DEBUG, "main() - init_uid() completed" );

    init_socket( sockpath );

    syslog( LOG_DEBUG, "main() - init_socket() completed" );

//  openlog( argv0, LOG_PID, LOG_MAIL );
    if( ! nodaemon )
	init_daemon();
    init_pidfile( pidfile );

    syslog( LOG_NOTICE, "%s %s compiled %s with %s starting", SPFMILTER_PROGRAM, SPFMILTER_VERSION, SPFMILTER_CURRENT, lib_version() );
    if( smfi_main() == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_main() failed" );
	exit( 1 );
	}
    syslog( LOG_NOTICE, "%s %s with %s terminating", SPFMILTER_PROGRAM, SPFMILTER_VERSION, lib_version() );

    fini_pidfile( pidfile );
    fini_socket( sockpath );
    fini_whitelist();
    fini_fallback();
    lib_fini();
    if( local_hostname != (char*) 0 )
	free( (void*) local_hostname );

    closelog();

    return 0;
}


static void
usage( void )
    {
    (void) fprintf( stderr, "usage: %s [options] /path/to/socket\n", argv0 );
    DOC_OPT( 'l', "localpolicy", "<spf-mechanism>", "Add local SPF policy mechanisms.", 5, 4 );
    DOC_OPT( 't', "trustedforwarders", (char*) 0, "Use the trusted-forwarder.org whitelist.", 20, 14 );
    DOC_OPT( 'g', "guess", "<spf-mechanisms>", "Add guess SPF policy mechanisms.", 4, 4 );
    DOC_OPT( 'f', "fallback", "<filename>", "A file of fallback SPF mechanisms.", 10, 14 );
    DOC_OPT( 'w', "whitelist", "<filename>", "A file of IP addresses to always accept mail from.", 10, 14 );
    DOC_OPT( 'r', "recipientmx", (char*) 0, "Check the MX records of the message's recipients (not implemented).", 20, 14 );
    DOC_OPT( 'e', "explanation", "<exp>", "Change the message returned in mail bounces.", 15, 14 );
    DOC_OPT( 'm', "markonly", (char*) 0, "Mark email instead of rejecting it.", 20, 22 );
    DOC_OPT( 'u', "user", "<user|uid>", "Run as specified user or UID.", 10, 16 );
    DOC_OPT( 'p', "pidfile", "<filename>", "Write the process i.d. to the specified file.", 10, 16 );
    DOC_OPT( 'X', "nodaemon", (char*) 0, "Do not fork into the background.", 20, 26 );
    DOC_OPT( 'h', "help", (char*) 0, "Show this help.", 20, 26 );
    DOC_OPT( 'd', "debug", "[<int>]", "Enable debugging to syslog.", 13, 18 );
    }


static void
init_fallback( const char* fallback_filename, const char* guess_str )
    {
    FILE* fp;
    char line[1000];
    char* cp;

    max_fallbacks = num_fallbacks = 0;
    fallbacks = (fallback_t*) 0;

    if( fallback_filename != (char*) 0 )
	{
	fp = fopen( fallback_filename, "r" );
	if( fp == (FILE*) 0 )
	    {
	    perror( fallback_filename );
	    exit( 1 );
	    }
	while ( fgets( line, sizeof(line), fp ) != (char*) 0 )
	    {
	    trim( line );
	    if( line[0] == '\0' )
		continue;
	    cp = line;
	    cp += strcspn( cp, " \t" );
	    *cp = '\0';
	    ++cp;
	    cp += strspn( cp, " \t" );
	    add_fallback( line, cp );
	    }
	(void) fclose( fp );
	}

    if( guess_str != (char*) 0 )
	add_fallback( "*", guess_str );
    }


static void
add_fallback( const char* pattern, const char* str )
    {
    if( num_fallbacks >= max_fallbacks )
	{
	if( max_fallbacks == 0 )
	    {
	    max_fallbacks = 50;		/* arbitrary */
	    fallbacks = (fallback_t*) malloc( max_fallbacks * sizeof(fallback_t) );
	    }
	else
	    {
	    max_fallbacks *= 2;
	    fallbacks = (fallback_t*) realloc( (void*) fallbacks, max_fallbacks * sizeof(fallback_t) );
	    }
	if( fallbacks == (fallback_t*) 0 )
	    {
	    (void) fprintf( stderr, "%s: out of memory enlarging fallbacks array\n", argv0 );
	    exit( 1 );
	    }
	}
    fallbacks[num_fallbacks].pattern = strdup( pattern );
    fallbacks[num_fallbacks].str = strdup( str );
    if( fallbacks[num_fallbacks].pattern == (char*) 0 ||
         fallbacks[num_fallbacks].str == (char*) 0 )
	{
	(void) fprintf( stderr, "%s: out of memory saving a fallback entry\n", argv0 );
	exit( 1 );
	}

    fallbacks[num_fallbacks].lib_fallback = lib_init_fallback( fallbacks[num_fallbacks].str );
    if( fallbacks[num_fallbacks].lib_fallback == (lib_fallback_t*) 0 )
	{
	(void) fprintf( stderr, "%s: error compiling fallback mechanism '%s'\n", argv0, fallbacks[num_fallbacks].str );
	exit( 1 );
	}

    ++num_fallbacks;
    }


static void
fini_fallback( void )
    {
    if( fallbacks != (fallback_t*) 0 )
	{
	int i;

	for( i = 0; i < num_fallbacks; ++i )
	    {
	    free( (void*) fallbacks[i].pattern );
	    free( (void*) fallbacks[i].str );
	    lib_fini_fallback( fallbacks[i].lib_fallback );
	    }
	max_fallbacks = num_fallbacks = 0;
	free( (void*) fallbacks );
	fallbacks = (fallback_t*) 0;
	}
    }


static int
find_fallback( const char* from )
    {
    const char* from_host;
    int i;

    from_host = strchr( from, '@' );
    if( from_host == (char*) 0 )
	return -1;
    ++from_host;
    if( fallbacks == (fallback_t*) 0 )
	return -1;
    for( i = 0; i < num_fallbacks; ++i )
	if( match( fallbacks[i].pattern, from_host ) )
	    return i;
    return -1;
    }


static void
init_whitelist( const char* whitelist_filename )
    {
    FILE* fp;
    char line[500];
    octets ip;

    if( whitelist_filename != (char*) 0 )
	{
	whitelist = iparray_new();
	if( whitelist == (iparray) 0 )
	    {
	    (void) fprintf( stderr, "%s: whitelist create failed\n", argv0 );
	    exit( 1 );
	    }
	fp = fopen( whitelist_filename, "r" );
	if( fp == (FILE*) 0 )
	    {
	    perror( whitelist_filename );
	    exit( 1 );
	    }
	while ( fgets( line, sizeof(line), fp ) != (char*) 0 )
	    {
	    trim( line );
	    if( line[0] == '\0' )
		continue;
	    if( iparray_parse_octets( line, &ip ) )
		(void) iparray_incr( whitelist, ip );
	    else
		(void) fprintf( stderr, "%s: unparsable IP address - \"%s\"\n", argv0, line );
	    }
	(void) fclose( fp );
	}
    else
	whitelist = (iparray) 0;
    }


static void
fini_whitelist( void )
    {
    syslog( LOG_DEBUG, "fini_whitelist()" );
    if( whitelist != (iparray) 0 )
	{
	iparray_delete( whitelist );
	iparray_fini();
	}
    }


static void
trim( char* str )
    {
    char* cp;
    int len;
    syslog( LOG_DEBUG, "trim(): [%s]", str );
    cp = strchr( str, '#' );
    if( cp != (char*) 0 )
	*cp = '\0';
    len = strlen( str );
    while ( str[len-1] == '\n' || str[len-1] == '\r' || str[len-1] == ' ' || str[len-1] == '\t' )
	{
	--len;
	str[len] = '\0';
	}
    }


static void
init_uid( const char* user )
    {
    struct passwd* pwd;
    char* ep;
    int uid;

    if( getuid() == 0 )
	{
        syslog( LOG_DEBUG, "init_uid(): Dropping privileges" );
	/* If we're root, the very first thing we do is become another user. */
	if( user == (char*) 0 )
	    (void) fprintf( stderr, "%s: warning: started as root but no --user flag specified\n", argv0 );
	else
	    {
	    /* Is it a number? */
	    uid = strtol( user, &ep, 0 );
	    if( *ep == '\0' )
		pwd = getpwuid( uid );
	    else
		pwd = getpwnam( user );
	    if( pwd == (struct passwd*) 0 )
		{
		(void) fprintf( stderr, "%s: unknown user: '%s'\n", argv0, user );
		exit( 1 );
		}
	    /* Set aux groups to null. */
	    if( setgroups( 0, (gid_t*) 0 ) < 0 )
		{
		perror( "setgroups" );
		exit( 1 );
		}
	    /* Set primary group. */
	    if( setgid( pwd->pw_gid ) < 0 )
		{
		perror( "setgid" );
		exit( 1 );
		}

	    /* Try setting aux groups correctly - not critical if this fails. */
	    if( initgroups( user, pwd->pw_gid ) < 0 )
		    perror( "initgroups" );
	    /* Set uid. */
	    if( setuid( pwd->pw_uid ) < 0 )
		{
		perror( "setuid" );
		exit( 1 );
		}
	    }
	}
    else
	{
	/* If we're not root but got a -user flag anyway, that's an error. */
	if( user != (char*) 0 )
	    {
	    (void) fprintf( stderr, "%s: can't switch users if not started as root\n", argv0 );
	    exit( 1 );
	    }
	}
    }


static void
init_socket( const char* sockpath )
    {
    unlink_socket( sockpath );

    /* Harden our umask so that the new socket gets created securely. */
    umask( 0077 );

    /* Initialize milter stuff. */
    syslog( LOG_DEBUG, "init_socket(): Registering milter" );
    smfi_setconn( (char*) sockpath );
    if( smfi_register( smfilter ) == MI_FAILURE )
	{
	(void) fprintf( stderr, "%s: smfi_register() failed\n", argv0 );
	exit( 1 );
	}
    }


static void
fini_socket( const char* sockpath )
    {
    unlink_socket( sockpath );
    }


static void
unlink_socket( const char* sockpath )
    {
    syslog( LOG_DEBUG, "unlink_socket(): Looking for local socket" );
    /* Remove old local socket.  Actually we should probably stat
    ** the file and check that it is in fact a socket before removing
    ** it.
    */
    if( strncasecmp( sockpath, "unix:", 5 ) == 0 )
	{
        syslog( LOG_DEBUG, "unlink_socket(): Removing local socket (a)" );
	if( unlink( &sockpath[5] ) < 0 )
	    if( errno != ENOENT )
		perror( &sockpath[5] );
	}
    else if( strncasecmp( sockpath, "local:", 6 ) == 0 )
	{
        syslog( LOG_DEBUG, "unlink_socket(): Removing local socket (b)" );
	if( unlink( &sockpath[6] ) < 0 )
	    if( errno != ENOENT )
		perror( &sockpath[6] );
	}
    else if( strchr( sockpath, ':' ) == (char*) 0 )
	{
        syslog( LOG_DEBUG, "unlink_socket(): Removing local socket (c)" );
	if( unlink( sockpath ) < 0 )
	    if( errno != ENOENT )
		perror( sockpath );
	}
    }


static void
init_daemon( void )
    {
    syslog( LOG_DEBUG, "init_daemon(): Daemonizing..." );
    /* Daemonize. */
#ifdef HAVE_DAEMON
    if( daemon( 0, 0 ) < 0)
	{
	perror( "daemon" );
	exit( 1 );
	}
#else /* HAVE_DAEMON */
    switch ( fork() )
	{
	case 0:
	    break;  
	case -1:
	    syslog( LOG_CRIT, "fork: %m" );
	    perror( "fork" );
	    exit( 1 );
	default:
	    exit( 0 );
	}
#ifdef HAVE_SETSID
    setsid();
#endif /* HAVE_SETSID */
#endif /* HAVE_DAEMON */
    }


static void
init_pidfile( const char* pidfile )
    {
    syslog( LOG_DEBUG, "init_pidfile(): Writing PID file" );
    if( pidfile != (char*) 0 )
	{
	FILE* fp;

	fp = fopen( pidfile, "w" );
	if( fp == (FILE*) 0 )
	    syslog( LOG_ERR, "unable to write PID file - %m" );
	else
	    {
	    (void) fprintf( fp, "%ld\n", (long) getpid() );
	    (void) fclose( fp );
	    }
	}
    }


static void
fini_pidfile( const char* pidfile )
    {
    syslog( LOG_DEBUG, "fini_pidfile(): Unlinking PID file" );
    if( pidfile != (char*) 0 )
	(void) unlink( pidfile );
    }


/* Milter routines. */


/* spf_connect - handle the initial TCP connection
**
** Called from: The sendmail process which wants us to opine on the SPF state of the connection.
** Called at the start of a connection.  Any per-connection data should be initialized here.
**
** connhost: The hostname of the client, based on a reverse lookup.
** connaddr: The client's IP address, based on getpeername().
*/
static sfsistat
spf_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr )
{
    connection_data_t* cd;
    struct sockaddr_in* sin;
    struct sockaddr_in6* sin6;
    const char* r;

    syslog( LOG_DEBUG, "spf_connect(): ============================================================ function entry" );

    if( connaddr == (_SOCK_ADDR*) 0 )
	return SMFIS_ACCEPT;	/* can't deal with it */

    syslog( LOG_DEBUG, "spf_connect(): initializing connection data" );

    /* Create & initialize the connection_data object. */
    cd = init_connection_data();

    syslog( LOG_DEBUG, "spf_connect(): connection data initialized" );

    if( cd == (connection_data_t*) 0 )
	{
	syslog( LOG_ERR, "couldn't allocate connection_data" );
	return SMFIS_TEMPFAIL;
	}
    if( smfi_setpriv( ctx, (void*) cd ) != MI_SUCCESS )
	{
	syslog( LOG_ERR, "smfi_setpriv() failed" );
	fini_connection_data( cd );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_connect(): smfi_setpriv completed" );

    /* The first time through, we figure out the local hostname. */
    /* We know if it's the first time through because, well, we don't know the local hostname yet. */
    /* And since we're only going to do this once per daemon, we can be reasonably relaxed about strdup()ing and free()ing memory, and */
    /* somewhat sanguine about that 200-character array on the stack... :) */
    /* If we can't figure out the local hostname we'll return SMFIS_TEMPFAIL to sendmail and quit trying. */
    if( local_hostname == (char*) 0 )
    {

        syslog( LOG_DEBUG, "spf_connect(): setting local hostname" );

	char* j_macro;	/* Sendmail's {j} macro - don't you just love Sendmail? */
	j_macro = smfi_getsymval( ctx, "{j}" );
	if( j_macro != (char*) 0 )
	    local_hostname = strdup( j_macro );
	else
	{
	    static char lh[200];
	    if( gethostname( lh, sizeof(lh) ) != 0 ) { syslog( LOG_ERR, "couldn't get local hostname" ); return SMFIS_TEMPFAIL; }
	    local_hostname = strdup( lh );
	}
	if( local_hostname == (char*) 0 ) { syslog( LOG_ERR, "couldn't strdup local hostname" ); return SMFIS_TEMPFAIL; }
	local_hostname_len = strlen( local_hostname );
	/* We'd better be careful with this call - it might allocate some memory and in that case we'll want to free it when the connection closes. */
	if( ! lib_set_local_hostname( cd->lib_data ) ) { syslog( LOG_ERR, "lib_set_local_hostname() failed" ); return SMFIS_TEMPFAIL; }

        syslog( LOG_DEBUG, "spf_connect(): setting local hostname - done" );

    }

    syslog( LOG_DEBUG, "spf_connect(): setting IP address" );

    /* Set the IP address. */
    switch ( connaddr->sa_family )
	{
	case AF_INET:

            syslog( LOG_DEBUG, "spf_connect(): it's an IPV4 address" );

	    sin = (struct sockaddr_in*) connaddr;
	    if( whitelist != (iparray) 0 )
		{
		char* char_addr = (char*) &sin->sin_addr.s_addr;
		octets ip;

		ip.a = char_addr[0];
		ip.b = char_addr[1];
		ip.c = char_addr[2];
		ip.d = char_addr[3];
		ip.w = 32;
		if( iparray_get( whitelist, ip ) > 0 )
		    {
		    syslog( LOG_INFO, "whitelist \"%s\" [%d.%d.%d.%d]", connhost, (int) ip.a, (int) ip.b, (int) ip.c, (int) ip.d );
		    cd->whitelisted = 1;
		    }
		}
	    r = inet_ntop( AF_INET, &sin->sin_addr, cd->buf1, sizeof(cd->buf1) );
	    if( r == (char*) 0 )
		{
		syslog( LOG_ERR, "couldn't convert IPv4 address to a string - %m" );
		return SMFIS_TEMPFAIL;
		}
	    if( ! lib_set_ipv4( cd->lib_data, sin->sin_addr, cd->buf1 ) )
		{
		syslog( LOG_ERR, "lib_set_ipv4() failed" );
		return SMFIS_TEMPFAIL;
		}
	    break;
	case AF_INET6:
	    sin6 = (struct sockaddr_in6*) connaddr;
	    r = inet_ntop( AF_INET6, &sin6->sin6_addr, cd->buf1, sizeof(cd->buf1) );
	    if( r == (char*) 0 )
		{
		syslog( LOG_ERR, "couldn't convert IPv6 address to a string - %m" );
		return SMFIS_TEMPFAIL;
		}
	    if( ! lib_set_ipv6( cd->lib_data, sin6->sin6_addr, cd->buf1 ) )
		{
		syslog( LOG_ERR, "lib_set_ipv6() failed" );
		return SMFIS_TEMPFAIL;
		}
	    break;
	default:
	    syslog( LOG_ERR, "unknown address family" );
	    return SMFIS_TEMPFAIL;
	}

        syslog( LOG_DEBUG, "spf_connect(): setting IP address - done - IP is %s", cd->buf1 );

    cd->ip_str = strdup( cd->buf1 );
    if( cd->ip_str == (char*) 0 )
    {
	syslog( LOG_ERR, "couldn't strdup ip_str" );
	return SMFIS_TEMPFAIL;
    }

    syslog( LOG_DEBUG, "spf_connect(): strdup() IP address - done - IP is %s", cd->ip_str );

    return SMFIS_CONTINUE;
}


/* spf_helo - handle the MAIL HELO command
**
** Called at the start of a connection.
**
** helohost: The string passed to the HELO/EHLO command.
*/
static sfsistat
spf_helo( SMFICTX* ctx, char* helohost )
    {
    connection_data_t* cd;

    syslog( LOG_DEBUG, "spf_helo(): [%s]", helohost );

    cd = (connection_data_t*) smfi_getpriv( ctx );

    syslog( LOG_DEBUG, "spf_helo(): smfi_getpriv() completed" );

    if( cd->helo != (char*) 0 )
	syslog( LOG_NOTICE, "multiple HELOs from %s [%s] - ignoring '%s'", cd->helo, cd->ip_str, helohost );
    else
	{
	cd->helo = strdup( helohost );
	if( cd->helo == (char*) 0 )
	    {
	    syslog( LOG_ERR, "couldn't strdup helo" );
	    return SMFIS_TEMPFAIL;
	    }
	}

    syslog( LOG_DEBUG, "spf_helo(): strdup() cd->helo=[%s]", cd->helo );

    if( cd->whitelisted )
	return SMFIS_CONTINUE;

    syslog( LOG_DEBUG, "spf_helo(): client is not whitelisted" );

    if( ! lib_set_helo_hostname( cd->lib_data, helohost ) )
	{
	syslog( LOG_ERR, "lib_set_helo_hostname() failed" );
	smfi_setreply( ctx, "452", "4.3.1", "Please try again later - lib_set_helo_hostname() failed." );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_helo(): lib_set_helo_hostname() succeeded" );

    return SMFIS_CONTINUE;
    }


/* spf_envfrom - handle the MAIL FROM:<> command
**
** Called at the start of each message.  There may be multiple messages
** in a connection.  Any per-message data should be initialized here.
**
** The sender address is also known as the "envelope from" address.
**
** fromargs: Null-terminated SMTP command arguments; fromargs[0] is the
**		   sender address.
*/
static sfsistat
spf_envfrom( SMFICTX* ctx, char** fromargs )
{
    connection_data_t* cd;
    char* verify;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    syslog( LOG_DEBUG, "spf_envfrom(): smfi_getpriv() completed" );

    if( ! init_message_data( cd ) )
	{
	syslog( LOG_ERR, "init_message_data() failed" );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_envfrom(): init_message_data() completed" );

    cd->mail_from = strdup( trim_address( fromargs[0] ) );
    if( cd->mail_from == (char*) 0 )
	{
	syslog( LOG_ERR, "couldn't strdup mail_from" );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_envfrom(): cd->mail_from=[%s]", cd->mail_from );

    if( cd->whitelisted )
	return SMFIS_CONTINUE;

    syslog( LOG_DEBUG, "spf_envfrom(): client is not whitelisted" );

    /* Authenticated messages get accepted without question. */
    verify = smfi_getsymval( ctx, "{verify}" );
    if( smfi_getsymval( ctx, "{auth_type}" ) != (char*) 0 ||
	 ( verify != (char*) 0 && strcmp( verify, "OK" ) == 0 ) )
	{
	cd->authenticated = 1;
	return SMFIS_CONTINUE;
	}

    syslog( LOG_DEBUG, "spf_envfrom(): client is not authenticated" );

    /* If the mail-from is empty or doesn't have an @, we make a fake one
    ** using the HELO domain (or if that's missing, the IP address).
    ** It's possible the SPF library we're using already does this, but at
    ** least some versions of some of the libraries do not, so we might
    ** as well do it here.
    */
    if( cd->mail_from[0] == '\0' || strchr( cd->mail_from, '@' ) == (char*) 0 )
	{
	if( cd->helo != (char*) 0 )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "mailer_daemon@%s", cd->helo );
	else
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "mailer_daemon@[%s]", cd->ip_str );
	cd->use_mail_from = strdup( cd->buf1 );
	}
    else
	cd->use_mail_from = strdup( cd->mail_from );
    if( cd->use_mail_from == (char*) 0 )
	{
	syslog( LOG_ERR, "couldn't strdup use_mail_from" );
	return SMFIS_TEMPFAIL;
	}

    if( ! lib_set_from( cd->lib_data, cd->use_mail_from ) )
	{
	syslog( LOG_ERR, "lib_set_from() failed" );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_envfrom(): cd->use_mail_from=[%s]", cd->use_mail_from );

    /* In MX mode we do a check for each recipient, and only later do the
    ** check for the whole message.
    */
    if( recipientmx )
	return SMFIS_CONTINUE;

    syslog( LOG_DEBUG, "spf_envfrom(): recipientmx is FALSE" );

    if( ! lib_do_check_m( cd->lib_data, cd->use_mail_from ) )
	{
	syslog( LOG_ERR, "lib_do_check_m() failed" );
	return SMFIS_TEMPFAIL;
	}

    syslog( LOG_DEBUG, "spf_envfrom(): lib_do_check_m returned.  Returned what?" );

//////////////////////////////////////////////////////////////////////////////////////// return SMFIS_ACCEPT;	/* We lied. */

    cd->result = lib_get_result( cd->lib_data );

    return handle_result( ctx );
}


/* spf_envrcpt - handle a RCPT TO:<> command
**
** Called separately for each recipient of a message.
**
** rcptargs: Null-terminated SMTP command arguments; rcptargs[0] is the
**           recipient address.
*/
static sfsistat
spf_envrcpt( SMFICTX* ctx, char** rcptargs )
    {
    connection_data_t* cd;
    char* rcpt;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if( cd->whitelisted || cd->authenticated )
	return SMFIS_CONTINUE;

    /* If we're not in MX mode then we've already checked the message. */
    if( ! recipientmx )
	return SMFIS_CONTINUE;

    rcpt = trim_address( rcptargs[0] );

    if( ! lib_do_check_recipient( cd->lib_data, rcpt ) )
	{
	syslog( LOG_ERR, "lib_do_check_recipient() failed" );
	return SMFIS_TEMPFAIL;
	}
    cd->result = lib_get_result( cd->lib_data );

    return handle_result( ctx );
    }


/* spf_header - handle a header line
**
** Called separately for each header line in a message.
**
** name:  Header field name.
** value: Header vield value, including folded whitespace.  The final CRLF
**		is removed.
*/
static sfsistat
spf_header( SMFICTX* ctx, char* name, char* value )
    {
    connection_data_t* cd;
    
    cd = (connection_data_t*) smfi_getpriv( ctx );

    /* We look for any instances of our own header in the incoming
    ** message, so we can remove them later.
    */
    if( strcasecmp( name, HEADER_NAME ) == 0 )
	{
	++cd->n_spf_headers;

	/* Does this SPF header claim to be from us? */
	if( receiver_is_me( value ) )
	    {
	    syslog( LOG_NOTICE, "found possible spoofed spf header - '%s'", value );
	    if( cd->n_del_headers < DEL_HEADER_MAX )
		{
		cd->del_header_list[cd->n_del_headers] = cd->n_spf_headers;
		++cd->n_del_headers;
		}
	    }
	}

    return SMFIS_CONTINUE;
    }


/* spf_eoh - handle the end of the headers
**
** Called once per message after all headers have been processed.
*/
static sfsistat
spf_eoh( SMFICTX* ctx )
    {
    connection_data_t* cd;
    
    cd = (connection_data_t*) smfi_getpriv( ctx );

    if( cd->whitelisted || cd->authenticated )
	return SMFIS_CONTINUE;

    if( ! recipientmx )
	return SMFIS_CONTINUE;

    if( ! lib_do_check_final( cd->lib_data ) )
	{
	syslog( LOG_ERR, "lib_do_check_final() failed" );
	return SMFIS_TEMPFAIL;
	}
    cd->result = lib_get_result( cd->lib_data );

    cd->action = SPFMILTER_ACTION_MARK;
    return SMFIS_CONTINUE;
    }


/* spf_eom - handle the end of the message
**
** Called once per message after all body blocks have been processed.
** Any per-message data should be freed both here and in spf_abort(),
** and wherever you return any of SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
spf_eom( SMFICTX* ctx )
    {
    connection_data_t* cd;
    int i;


    cd = (connection_data_t*) smfi_getpriv( ctx );

    /* Header deleting and adding can only happen in the eom handler. */

    /* Remove any previous SPF headers, to prevent spoofing. */
    if( cd->n_del_headers >= DEL_HEADER_MAX )
	{
	/* There were too many bogus SPF headers to keep track of, so to
	** be safe we just delete all SPF headers.
	*/
	for( i = cd->n_spf_headers; i >= 1; --i )
	    smfi_chgheader( ctx, HEADER_NAME, i, (char*) 0 );
	}
    else
	{
	/* We have a list of only the SPF headers purporting to be from us;
	** those get deleted.  We do the deletions in reverse order in case
	** deleting one header causes subsequent headers to get re-numbered.
	** (I actually checked whether this happens in the current milter
	** implementation, and it does NOT, but perhaps some future version
	** will get it wrong.)
	*/
	for( i = cd->n_del_headers - 1 ; i >= 0; --i )
	    smfi_chgheader( ctx, HEADER_NAME, cd->del_header_list[i], (char*) 0 );
	}

    if( cd->whitelisted || cd->authenticated )
	{
	cd->result = SPFMILTER_RESULT_PASS;
	cd->action = SPFMILTER_ACTION_MARK;
	if( cd->whitelisted )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "%s: %s is whitelisted", local_hostname, cd->ip_str );
	else if( cd->authenticated )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "%s: authenticated connection", local_hostname );
	else
	    (void) strncpy( cd->buf1, "???", sizeof(cd->buf1) );

	}
    else
	reason_str( cd, cd->buf1, sizeof(cd->buf1) );

    if( cd->action == SPFMILTER_ACTION_MARK )
	{
	/* Build and add the SPF header. */
	build_header( cd, cd->buf2, sizeof(cd->buf2), cd->buf1 );
	/* The SPF spec wants us to prepend the new header to the existing ones, but not all versions of the milter API have the necessary function. */
#ifdef HAVE_SMFI_INSHEADER
	smfi_insheader( ctx, 0, HEADER_NAME, cd->buf2 );
#else
	smfi_addheader( ctx, HEADER_NAME, cd->buf2 );
#endif
	}

    fini_message_data( cd );

    return SMFIS_CONTINUE;
    }


/* spf_abort - handle the message being aborted
**
** May be called at any time during processing of a message, indicating
** that something went wrong.  Any per-message data should be freed both
** here and in sample_eom(), and wherever you return any of
** SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
spf_abort( SMFICTX* ctx )
{
    connection_data_t* cd;
    cd = (connection_data_t*) smfi_getpriv( ctx );
    if( cd == (connection_data_t*) 0 ) { syslog( LOG_NOTICE, "spf_abort called but cd is null" ); }
    else
    {
	fini_message_data( cd );
    }
    return SMFIS_CONTINUE;
}


/* spf_close - handle the connection being closed
**
** Called once at the end of a connection.  Any per-connection data
** should be freed here.
*/
static sfsistat
spf_close( SMFICTX* ctx )
    {
    connection_data_t* cd;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if( cd != (connection_data_t*) 0 )
	{
	smfi_setpriv( ctx, (void*) 0 );
	fini_connection_data( cd );
	}
    else
	syslog( LOG_NOTICE, "spf_close called but cd is null" );

    return SMFIS_CONTINUE;
    }


static char*
trim_address( char* address )
    {
    int len;
    syslog( LOG_DEBUG, "trim_address(): [%s]", address );
    /* The milter API sometimes gives us strings with angle brackets. */
    while ( *address == '<' )
	++address;
    len = strlen( address );
    while ( len > 0 && address[len - 1] == '>' )
	address[--len] = '\0';

    /* We also sometimes get CRs on the ends of addresses. */
    while ( len > 0 && address[len - 1] == '\015' )
	address[--len] = '\0';
    syslog( LOG_DEBUG, "trim_address(): [%s] - done", address );
    return address;
    }


static sfsistat
handle_result( SMFICTX* ctx )
    {
    connection_data_t* cd;
    const char* exp;
    char exp_escaped[1000];

    cd = (connection_data_t*) smfi_getpriv( ctx );

    switch ( cd->result )
	{
	case SPFMILTER_RESULT_PASS:
	case SPFMILTER_RESULT_SOFTFAIL:
	case SPFMILTER_RESULT_NEUTRAL:
	case SPFMILTER_RESULT_UNKNOWN:
	case SPFMILTER_RESULT_UNMECH:
	case SPFMILTER_RESULT_ERROR:	/* typically "DNS lookup failure" */
	case SPFMILTER_RESULT_NONE:
	    cd->action = SPFMILTER_ACTION_MARK;
	    break;
	case SPFMILTER_RESULT_FAIL:
	    if( markonly )
		cd->action = SPFMILTER_ACTION_MARK;
	    else
		cd->action = SPFMILTER_ACTION_REJECT;
	    break;
	}

    switch ( cd->action )
	{
	case SPFMILTER_ACTION_REJECT:
	    exp = lib_get_explanation( cd->lib_data );
	    syslog( LOG_INFO, "rejecting mail from [%s] - %s", cd->ip_str, result_str( cd->result ) );
	    if( exp != (char*) 0 )
		escape_percents( exp, exp_escaped, sizeof(exp_escaped) );
	    else
		(void) strncpy( exp_escaped, "rejected by spfmilter", sizeof(exp_escaped) - 1 );
	    smfi_setreply( ctx, "550", "5.7.1", exp_escaped );
	    fini_message_data( cd );
	    return SMFIS_REJECT;

	case SPFMILTER_ACTION_TEMPFAIL:
	    syslog( LOG_INFO, "temporarily failing mail from [%s] - %s", cd->ip_str, lib_get_error( cd->lib_data ) );
	    smfi_setreply( ctx, "451", "4.3.2", "Please try again later." );
	    fini_message_data( cd );
	    return SMFIS_TEMPFAIL;

	default:
	    return SMFIS_CONTINUE;
	}
    }


/* escape_percents - escape '%' characters in a string
**
** Sendmail doesn't like sending percent characters in bounce messages.
** If a milter returns an error message with percents in it, sendmail
** will ignore that and instead return an uninformative default
** message saying "Command rejected".  To avoid this we escape the
** percents witrh a second percent.
*/
static void
escape_percents( const char* in, char* out, int out_size )
    {
    const char* i;
    char* o;
    int out_len;

    i = in;
    o = out;
    out_len = 0;
    while ( *i != '\0' && out_len + 2 < out_size )
	{
	if( *i == '%' )
	    {
	    /* The character that most often comes back escaped is
	    ** the @ sign (%40). Since it doesn't actually need
	    ** escaping, unescape it.
	    */
	    if( i[1] == '4' && i[2] == '0' )
		{
		*o++ = '@';
		++out_len;
		i += 3;
		}
	    else
		{
		/* Otherwise, escape the %-sign. */
		*o++ = '%';
		*o++ = '%';
		out_len += 2;
		++i;
		}
	    }
	else
	    {
	    *o++ = *i++;
	    ++out_len;
	    }
	}
    *o = '\0';
    }


static int
receiver_is_me( const char* value )
    {
    const char* cp;

    /* First look for a receiver= setting after the comment part.
    ** The hostname immediately follows the equals and is immediately
    ** followed by a semicolon.
    */
    cp = strrchr( value, ')' );
    if( cp != (char*) 0 )
	++cp;
    else
	cp = value;
    cp = strstr( cp, " receiver=" );
    if( cp != (char*) 0 )
	{
	cp += 10;
	if( strncasecmp( cp, local_hostname, local_hostname_len ) == 0 )
	    {
	    cp += local_hostname_len;
	    if( *cp == ';' )
		return 1;	/* Yes! */
	    }
	}

    /* If there was no receiver= (it's optional), try checking the
    ** comment part for a hostname.  That's even more optional
    ** (it's not in the spec at all), but many implementations do add it.
    **
    ** To find the comment hostname, skip one word, whitespace, and
    ** an open-paren, then check for the hostname followed by a colon.
    */
    cp = value;
    cp += strcspn( cp, " \t" );
    cp += strspn( cp, " \t" );
    if( *cp == '(' )
	{
	++cp;
	if( strncasecmp( cp, local_hostname, local_hostname_len ) == 0 )
	    {
	    cp += local_hostname_len;
	    if( *cp == ':' )
		return 1;	/* Yes! */
	    }
	}

    /* Nope. */
    return 0;
    }


static void
build_header( connection_data_t* cd, char* header, size_t header_size, const char* reason )
    {
    int len;

    (void) snprintf( header, header_size, "%s", result_str( cd->result ) );
    len = strlen( header );
    if( reason != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " (%s)", reason );
	len = strlen( header );
	}
    if( local_hostname != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " receiver=%s;", local_hostname );
	len = strlen( header );
	}
    if( cd->ip_str!= (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " client-ip=%s;", cd->ip_str);
	len = strlen( header );
	}
    if( cd->helo != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " helo=%s;", cd->helo );
	len = strlen( header );
	}
    if( cd->mail_from != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " envelope-from=%s;", cd->mail_from );
	len = strlen( header );
	}
    /*!!! Do something about the problem= field. */
    (void) snprintf( &header[len], header_size - len, " x-software=%s %s with %s;", SPFMILTER_PROGRAM, SPFMILTER_VERSION, lib_version() );
    }


// init_connection_data()
// Create a connection_data_s structure and initialize the contents to nothing much at all.
// The only thing in the structure that will not be nothing at this stage is the pointer to the
// lib_data_s structure.  That will contain a pointer to a newly-created SPF request object and
// a pointer to a SPF response object which for the moment will have been initialized to NULL.
// Return a pointer to our shiny new connection_data_s structure.
//
static connection_data_t*
init_connection_data( void )
{
    connection_data_t* cd;
    cd = (connection_data_t*) malloc( sizeof(connection_data_t) );
    if( cd == (connection_data_t*) 0 ) { return (connection_data_t*) 0; }
    cd->whitelisted = 0;
    cd->ip_str = (char*) 0;
    cd->helo = (char*) 0;
    cd->lib_data = lib_init_connection_data();
    if( cd->lib_data == (lib_data_t*) 0 ) { free( (void*) cd ); return(connection_data_t*) 0; }
    if( ! init_message_data( cd ) ) { fini_connection_data( cd ); return(connection_data_t*) 0; }
    return cd;
}


// init_message_data()
// Initialize per-message data.  Called at least once (at the time of creation of the per-connection data) but
// may additionally be called multiple times per connection as there may be several messages per connection.
// Called from: init_connection_data(), spf_envfrom()
// Returns: 1 on success, 0 on failure.
// 
static int
init_message_data( connection_data_t* cd )
{
    cd->authenticated = 0;
    cd->mail_from = (char*) 0;
    cd->use_mail_from = (char*) 0;
    cd->result = -1;
    cd->action = SPFMILTER_ACTION_UNKNOWN;
    cd->n_spf_headers = 0;
    cd->n_del_headers = 0;
    if( ! lib_init_message_data( cd->lib_data ) ) { return 0; }		// The library call returns 1 on success, 0 on failure.
    return 1;
}


/* Frees or resets any items in the per-connection data structure that are for an individual message. */
static void
fini_message_data( connection_data_t* cd )
{
    if( cd->mail_from     != (char*) 0 ) { free( (void*) cd->mail_from );     cd->mail_from     = (char*) 0; }
    if( cd->use_mail_from != (char*) 0 ) { free( (void*) cd->use_mail_from ); cd->use_mail_from = (char*) 0; }
    lib_fini_message_data( cd->lib_data );
}


/* Frees all items in the connection_data. */
static void
fini_connection_data( connection_data_t* cd )
{
    fini_message_data( cd );
    if( cd->ip_str != (char*) 0 ) { free( (void*) cd->ip_str ); }
    if( cd->helo   != (char*) 0 ) { free( (void*) cd->helo ); }
    lib_fini_connection_data( cd->lib_data );
    free( (void*) cd );
}


static char*
result_str( int result )
    {
    switch ( result )
	{
	case SPFMILTER_RESULT_INVALID: return "internal error";
	case SPFMILTER_RESULT_PASS: return "pass";
	case SPFMILTER_RESULT_FAIL: return "fail";
	case SPFMILTER_RESULT_SOFTFAIL: return "softfail";
	case SPFMILTER_RESULT_NEUTRAL: return "neutral";
	case SPFMILTER_RESULT_UNKNOWN: return "unknown";
	case SPFMILTER_RESULT_UNMECH: return "unknown-mechanisms";
	case SPFMILTER_RESULT_ERROR: return "error";
	case SPFMILTER_RESULT_NONE: return "none";
	default: return "???";
	}
    }


static void
reason_str( connection_data_t* cd, char* str, size_t size )
    {
    switch ( cd->result )
	{
	case SPFMILTER_RESULT_INVALID:
	(void) snprintf( str, size, "%s: internal error in processing during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_PASS:
	(void) snprintf( str, size, "%s: domain of %s designates %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_FAIL:
	(void) snprintf( str, size, "%s: domain of %s does not designate %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_SOFTFAIL:
	(void) snprintf( str, size, "%s: domain of transitioning %s does not designate %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_NEUTRAL:
	(void) snprintf( str, size, "%s: %s is neither permitted nor denied by domain of %s", local_hostname, cd->ip_str, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_UNKNOWN:
	(void) snprintf( str, size, "%s: unknown result during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_UNMECH:
	(void) snprintf( str, size, "%s: domain of %s uses a mechanism not recognized by this client", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_ERROR:
	(void) snprintf( str, size, "%s: error in processing during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_NONE:
	(void) snprintf( str, size, "%s: %s does not designate permitted sender hosts", local_hostname, cd->use_mail_from );
	break;
	default:
	(void) snprintf( str, size, "%s: ???", local_hostname );
	break;
	}
    }


#ifdef USE_LIBSPF

#error "This milter is not intended to use libspf - please use libspf2-1.2.9 or later."

/* Very old Libspf. */  #include <libspf/spf.h>
/* Very old Libspf. */  
/* Very old Libspf. */  struct lib_fallback_s {
/* Very old Libspf. */      int junk;	/* not used, but mallocing zero bytes can be a problem */
/* Very old Libspf. */      };
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  struct lib_data_s {
/* Very old Libspf. */      peer_info_t* peer_info;
/* Very old Libspf. */      SPF_RESULT res;
/* Very old Libspf. */      };
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_init( void )
/* Very old Libspf. */      {
/* Very old Libspf. */      if( recipientmx )
/* Very old Libspf. */  	{
/* Very old Libspf. */  	(void) fprintf( stderr, "Sorry, --recipientmx mode is not implemented with libspf; if you want" );
/* Very old Libspf. */  	(void) fprintf( stderr, "to use that flag you will need to rebuild spfmilter with libspf2.\n" );
/* Very old Libspf. */  	return 0;
/* Very old Libspf. */  	}
/* Very old Libspf. */  
/* Very old Libspf. */      /* SPF_dbg_level( debug ); */
/* Very old Libspf. */  
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static const char*
/* Very old Libspf. */  lib_version( void )
/* Very old Libspf. */      {
/* Very old Libspf. */      return "libspf-unknown";	/* the library doesn't tell us its version */
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static lib_fallback_t*
/* Very old Libspf. */  lib_init_fallback( const char* str )
/* Very old Libspf. */      {
/* Very old Libspf. */      lib_fallback_t* lf;
/* Very old Libspf. */  
/* Very old Libspf. */      lf = (lib_fallback_t*) malloc( sizeof(lib_fallback_t) );
/* Very old Libspf. */      if( lf == (lib_fallback_t*) 0 )
/* Very old Libspf. */  	return (lib_fallback_t*) 0;
/* Very old Libspf. */      return lf;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  /* Not implemented. */
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_set_local_hostname( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static lib_data_t*
/* Very old Libspf. */  lib_init_connection_data( void )
/* Very old Libspf. */      {
/* Very old Libspf. */      lib_data_t* ld;
/* Very old Libspf. */  
/* Very old Libspf. */      ld = (lib_data_t*) malloc( sizeof(lib_data_t) );
/* Very old Libspf. */      if( ld == (lib_data_t*) 0 )
/* Very old Libspf. */  	return (lib_data_t*) 0;
/* Very old Libspf. */      return ld;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_init_message_data( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str )
/* Very old Libspf. */      {
/* Very old Libspf. */      ld->peer_info = SPF_init( local_hostname, ipv4_str, explanation_str, (char*) 0, (char*) 0, trustedforwarders, 0 );
/* Very old Libspf. */      if( ld->peer_info == (peer_info_t*) 0 )
/* Very old Libspf. */  	return 0;
/* Very old Libspf. */      ld->res = -1;
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str )
/* Very old Libspf. */      {
/* Very old Libspf. */      ld->peer_info = SPF_init( local_hostname, ipv6_str, explanation_str, (char*) 0, (char*) 0, trustedforwarders, 0 );
/* Very old Libspf. */      if( ld->peer_info == (peer_info_t*) 0 )
/* Very old Libspf. */  	return 0;
/* Very old Libspf. */      ld->res = -1;
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname )
/* Very old Libspf. */      {
/* Very old Libspf. */      if( ! SPF_smtp_helo( ld->peer_info, helo_hostname ) )
/* Very old Libspf. */  	return 0;
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_set_from( lib_data_t* ld, const char* from )
/* Very old Libspf. */      {
/* Very old Libspf. */      if( ! SPF_smtp_from( ld->peer_info, from ) )
/* Very old Libspf. */  	return 0;
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_do_check_m( lib_data_t* ld, const char* from )
/* Very old Libspf. */      {
/* Very old Libspf. */      ld->res = SPF_policy_main( ld->peer_info );
/* Very old Libspf. */  
/* Very old Libspf. */      /* If there was no SPF record, then try finding a fallback mechanism. */
/* Very old Libspf. */      if( ld->res == SPF_NONE )
/* Very old Libspf. */  	{
/* Very old Libspf. */  	int i = find_fallback( from );
/* Very old Libspf. */  
/* Very old Libspf. */  	if( i != -1 )
/* Very old Libspf. */  	    {
/* Very old Libspf. */  	    if( ! SPF_parse_policy( ld->peer_info, fallbacks[i].str ) )
/* Very old Libspf. */  		return 0;
/* Very old Libspf. */  	    ld->res = ld->peer_info->RES;
/* Very old Libspf. */  	    }
/* Very old Libspf. */  	}
/* Very old Libspf. */  
/* Very old Libspf. */      return 1;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_do_check_recipient( lib_data_t* ld, const char* to )
/* Very old Libspf. */      {
/* Very old Libspf. */      /* Not implemented for libspf. */
/* Very old Libspf. */      return 0;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_do_check_final( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      /* Not implemented for libspf. */
/* Very old Libspf. */      return 0;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static int
/* Very old Libspf. */  lib_get_result( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      /* Convert libspf result to spfmilter result. */
/* Very old Libspf. */      switch ( ld->res )
/* Very old Libspf. */  	{
/* Very old Libspf. */  	case SPF_PASS: return SPFMILTER_RESULT_PASS;
/* Very old Libspf. */  	case SPF_NONE: return SPFMILTER_RESULT_NONE;
/* Very old Libspf. */  	case SPF_S_FAIL: return SPFMILTER_RESULT_SOFTFAIL;
/* Very old Libspf. */  	case SPF_H_FAIL: return SPFMILTER_RESULT_FAIL;
/* Very old Libspf. */  	case SPF_ERROR: return SPFMILTER_RESULT_ERROR;
/* Very old Libspf. */  	case SPF_NEUTRAL: return SPFMILTER_RESULT_NEUTRAL;
/* Very old Libspf. */  	case SPF_UNKNOWN: return SPFMILTER_RESULT_UNKNOWN;
/* Very old Libspf. */  	case SPF_UNMECH: return SPFMILTER_RESULT_UNMECH;
/* Very old Libspf. */  	default: return -1;
/* Very old Libspf. */  	}
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static const char*
/* Very old Libspf. */  lib_get_explanation( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      return ld->peer_info->explain;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static const char*
/* Very old Libspf. */  lib_get_error( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      return ld->peer_info->error;
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static void
/* Very old Libspf. */  lib_fini_message_data( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static void
/* Very old Libspf. */  lib_fini_connection_data( lib_data_t* ld )
/* Very old Libspf. */      {
/* Very old Libspf. */      if( ld != (lib_data_t*) 0 )
/* Very old Libspf. */  	{
/* Very old Libspf. */  	if( ld->peer_info != (peer_info_t*) 0 )
/* Very old Libspf. */  	    ld->peer_info = SPF_close( ld->peer_info );
/* Very old Libspf. */  	free( (void*) ld );
/* Very old Libspf. */  	}
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static void
/* Very old Libspf. */  lib_fini_fallback( lib_fallback_t* lf )
/* Very old Libspf. */      {
/* Very old Libspf. */      if( lf != (lib_fallback_t*) 0 )
/* Very old Libspf. */  	free( (void*) lf );
/* Very old Libspf. */      }
/* Very old Libspf. */  
/* Very old Libspf. */  
/* Very old Libspf. */  static void
/* Very old Libspf. */  lib_fini( void )
/* Very old Libspf. */      {
/* Very old Libspf. */      }

#endif /* USE_LIBSPF */

#ifdef USE_LIBSPF2

/* Libspf2 data and routines. */

#include <spf2/spf.h>

/* Which version of libspf2 do we have? */
#if( SPF_LIB_VERSION_MAJOR == 1 && SPF_LIB_VERSION_MINOR < 2 )
#error "This milter is not intended to use libspf2 versions before libspf2-1.2 - please use libspf2-1.2.9 or later."

/* Old libspf2 API. */
/* Old libspf2 API. */  
/* Old libspf2 API. */  #undef USE_DNS_CACHE
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  #include <spf2/spf_dns_resolv.h>
/* Old libspf2 API. */  #ifdef USE_DNS_CACHE
/* Old libspf2 API. */  #include <spf2/spf_dns_cache.h>
/* Old libspf2 API. */  #endif /* USE_DNS_CACHE */
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static SPF_config_t spfc_global;
/* Old libspf2 API. */  static SPF_c_results_t spfr_localpolicy;
/* Old libspf2 API. */  static SPF_c_results_t spfr_explanation;
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  struct lib_fallback_s {
/* Old libspf2 API. */      SPF_c_results_t results;
/* Old libspf2 API. */      };
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  struct lib_data_s {
/* Old libspf2 API. */      SPF_config_t spfcid;
/* Old libspf2 API. */      const char* from;
/* Old libspf2 API. */      SPF_output_t output;
/* Old libspf2 API. */      int got_output;
/* Old libspf2 API. */      };
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  /* The libspf2 code uses a resolver pool to work around a memory leak
/* Old libspf2 API. */  ** in the resolver library.
/* Old libspf2 API. */  */
/* Old libspf2 API. */  static void init_resolvers( void );
/* Old libspf2 API. */  static void fini_resolvers( void );
/* Old libspf2 API. */  static int find_resolver( void );
/* Old libspf2 API. */  static SPF_dns_config_t get_resolver( int r );
/* Old libspf2 API. */  static void free_resolver( int r );
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_init( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      spfc_global = SPF_create_config();
/* Old libspf2 API. */      if( spfc_global == (SPF_config_t) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	(void) fprintf( stderr, "%s: SPF_create_config() failed\n", argv0 );
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      SPF_set_debug( spfc_global, debug );
/* Old libspf2 API. */  
/* Old libspf2 API. */      if( localpolicy_str != (char*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_init_c_results( &spfr_localpolicy );
/* Old libspf2 API. */  	if( SPF_compile_local_policy( spfc_global, localpolicy_str, trustedforwarders, &spfr_localpolicy ) )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    (void) fprintf( stderr, "%s: error compiling local policy - %s\n", argv0, spfr_localpolicy.err_msg );
/* Old libspf2 API. */  	    return 0;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	SPF_set_local_policy( spfc_global, spfr_localpolicy );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      if( explanation_str != (char*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_init_c_results( &spfr_explanation );
/* Old libspf2 API. */  	if( SPF_compile_exp( spfc_global, explanation_str, &spfr_explanation ) )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    (void) fprintf( stderr, "%s: error compiling explanation - %s\n", argv0, spfr_explanation.err_msg );
/* Old libspf2 API. */  	    return 0;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	SPF_set_exp( spfc_global, spfr_explanation );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      init_resolvers();
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static const char*
/* Old libspf2 API. */  lib_version( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      static char version[100];
/* Old libspf2 API. */  
/* Old libspf2 API. */      (void) snprintf( version, sizeof(version), "libspf2-%d.%d.%d", SPF_LIB_VERSION_MAJOR, SPF_LIB_VERSION_MINOR, SPF_LIB_VERSION_PATCH );
/* Old libspf2 API. */      return version;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static lib_fallback_t*
/* Old libspf2 API. */  lib_init_fallback( const char* str )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      lib_fallback_t* lf;
/* Old libspf2 API. */  
/* Old libspf2 API. */      lf = (lib_fallback_t*) malloc( sizeof(lib_fallback_t) );
/* Old libspf2 API. */      if( lf == (lib_fallback_t*) 0 )
/* Old libspf2 API. */  	return (lib_fallback_t*) 0;
/* Old libspf2 API. */      SPF_init_c_results( &lf->results );
/* Old libspf2 API. */      if( SPF_compile_local_policy( spfc_global, str, trustedforwarders, &lf->results ) )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	free( (void*) lf );
/* Old libspf2 API. */  	return (lib_fallback_t*) 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      return lf;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_set_local_hostname( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( SPF_set_rec_dom( spfc_global, local_hostname ) )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static lib_data_t*
/* Old libspf2 API. */  lib_init_connection_data( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      lib_data_t* ld;
/* Old libspf2 API. */  
/* Old libspf2 API. */      ld = (lib_data_t*) malloc( sizeof(lib_data_t) );
/* Old libspf2 API. */      if( ld == (lib_data_t*) 0 )
/* Old libspf2 API. */  	return (lib_data_t*) 0;
/* Old libspf2 API. */      ld->from = (char*) 0;
/* Old libspf2 API. */      ld->got_output = 0;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Create the per-connection config object by duping the global one. */
/* Old libspf2 API. */      ld->spfcid = SPF_dup_config( spfc_global );
/* Old libspf2 API. */      if( ld->spfcid == (SPF_config_t) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	free( (void*) ld );
/* Old libspf2 API. */  	return (lib_data_t*) 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      return ld;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_init_message_data( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      ld->from = (char*) 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( SPF_set_ipv4( ld->spfcid, ipv4_addr ) )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( SPF_set_ipv6( ld->spfcid, ipv6_addr ) )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( SPF_set_helo_dom( ld->spfcid, helo_hostname ) )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_set_from( lib_data_t* ld, const char* from )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      ld->from = strdup( from );
/* Old libspf2 API. */      if( ld->from == (char*) 0 )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      if( SPF_set_env_from( ld->spfcid, ld->from ) )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_do_check_m( lib_data_t* ld, const char* from )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Find a resolver to use. */
/* Old libspf2 API. */      r  = find_resolver();
/* Old libspf2 API. */      if( r < 0 )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Do the SPF lookup. */
/* Old libspf2 API. */      ld->output = SPF_result( ld->spfcid, get_resolver( r ) );
/* Old libspf2 API. */      ld->got_output = 1;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* If there was no SPF record, then try finding a fallback mechanism. */
/* Old libspf2 API. */      if( ld->output.result == SPF_RESULT_NONE )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	int i = find_fallback( from );
/* Old libspf2 API. */  	if( i != -1 )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    SPF_free_output( &ld->output );
/* Old libspf2 API. */  	    ld->output = SPF_eval_id( ld->spfcid, fallbacks[i].lib_fallback->results.spfid, get_resolver( r ), 1, 0, (int*) 0 );
/* Old libspf2 API. */  	    SPF_result_comments( ld->spfcid, get_resolver( r ), fallbacks[i].lib_fallback->results, &ld->output );
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* And put the resolver back into the pool. */
/* Old libspf2 API. */      free_resolver( r );
/* Old libspf2 API. */  
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_do_check_recipient( lib_data_t* ld, const char* to )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Find a resolver to use. */
/* Old libspf2 API. */      r  = find_resolver();
/* Old libspf2 API. */      if( r < 0 )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* If we have already checked a previous recipient, discard that output, it's no longer needed. */
/* Old libspf2 API. */      if( ld->got_output )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_free_output( &ld->output );
/* Old libspf2 API. */  	ld->got_output = 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Do the SPF lookup. */
/* Old libspf2 API. */      ld->output = SPF_result_2mx( ld->spfcid, get_resolver( r ), to );
/* Old libspf2 API. */      ld->got_output = 1;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* If there was no SPF record, then try finding a fallback mechanism. */
/* Old libspf2 API. */      if( ld->output.result == SPF_RESULT_NONE )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	int i = find_fallback( ld->from );
/* Old libspf2 API. */  	if( i != -1 )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    SPF_free_output( &ld->output );
/* Old libspf2 API. */  	    ld->output = SPF_eval_id( ld->spfcid, fallbacks[i].lib_fallback->results.spfid, get_resolver( r ), 1, 0, (int*) 0 );
/* Old libspf2 API. */  	    SPF_result_comments( ld->spfcid, get_resolver( r ), fallbacks[i].lib_fallback->results, &ld->output );
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* And put the resolver back into the pool. */
/* Old libspf2 API. */      free_resolver( r );
/* Old libspf2 API. */  
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_do_check_final( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Find a resolver to use. */
/* Old libspf2 API. */      r  = find_resolver();
/* Old libspf2 API. */      if( r < 0 )
/* Old libspf2 API. */  	return 0;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Discard output from any recipient checks. */
/* Old libspf2 API. */      if( ld->got_output )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_free_output( &ld->output );
/* Old libspf2 API. */  	ld->got_output = 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Get the accumulated results for all recipients. */
/* Old libspf2 API. */      ld->output = SPF_result_2mx_msg( ld->spfcid, get_resolver( r ) );
/* Old libspf2 API. */      ld->got_output = 1;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* And put the resolver back into the pool. */
/* Old libspf2 API. */      free_resolver( r );
/* Old libspf2 API. */  
/* Old libspf2 API. */      return 1;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  lib_get_result( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      /* Convert libspf2 result to spfmilter result. */
/* Old libspf2 API. */      switch ( ld->output.result )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	case SPF_RESULT_NONE: return SPFMILTER_RESULT_NONE;
/* Old libspf2 API. */  	case SPF_RESULT_NEUTRAL: return SPFMILTER_RESULT_NEUTRAL;
/* Old libspf2 API. */  	case SPF_RESULT_PASS: return SPFMILTER_RESULT_PASS;
/* Old libspf2 API. */  	case SPF_RESULT_FAIL: return SPFMILTER_RESULT_FAIL;
/* Old libspf2 API. */  	case SPF_RESULT_SOFTFAIL: return SPFMILTER_RESULT_SOFTFAIL;
/* Old libspf2 API. */  	case SPF_RESULT_UNKNOWN: return SPFMILTER_RESULT_UNKNOWN;
/* Old libspf2 API. */  	case SPF_RESULT_ERROR: return SPFMILTER_RESULT_ERROR;
/* Old libspf2 API. */  	default: return -1;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static const char*
/* Old libspf2 API. */  lib_get_reason( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      return ld->output.header_comment;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static const char*
/* Old libspf2 API. */  lib_get_explanation( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      return ld->output.smtp_comment;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static const char*
/* Old libspf2 API. */  lib_get_error( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      return SPF_strerror( ld->output.err );
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  lib_fini_message_data( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( ld->from != (char*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	free( (void*) ld->from );
/* Old libspf2 API. */  	ld->from = (char*) 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      if( ld->got_output )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_free_output( &ld->output );
/* Old libspf2 API. */  	ld->got_output = 0;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  lib_fini_connection_data( lib_data_t* ld )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( ld != (lib_data_t*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	if( ld->spfcid != (SPF_config_t) 0 )
/* Old libspf2 API. */  	    SPF_destroy_config( ld->spfcid );
/* Old libspf2 API. */  	free( (void*) ld );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  lib_fini_fallback( lib_fallback_t* lf )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      if( lf != (lib_fallback_t*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	SPF_free_c_results( &lf->results );
/* Old libspf2 API. */  	free( (void*) lf );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  lib_fini( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      fini_resolvers();
/* Old libspf2 API. */      if( localpolicy_str != (char*) 0 )
/* Old libspf2 API. */  	SPF_free_c_results( &spfr_localpolicy );
/* Old libspf2 API. */      if( explanation_str != (char*) 0 )
/* Old libspf2 API. */  	SPF_free_c_results( &spfr_explanation );
/* Old libspf2 API. */      SPF_destroy_config( spfc_global );
/* Old libspf2 API. */      SPF_destroy_default_config();
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  /* Resolver pool. */
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  /* Resolver data. */
/* Old libspf2 API. */  typedef struct {
/* Old libspf2 API. */      int initialized, inuse;
/* Old libspf2 API. */      int next_free_resolver;
/* Old libspf2 API. */      SPF_dns_config_t spfdcid_r;
/* Old libspf2 API. */  #ifdef USE_DNS_CACHE
/* Old libspf2 API. */      SPF_dns_config_t spfdcid_c;
/* Old libspf2 API. */  #endif /* USE_DNS_CACHE */
/* Old libspf2 API. */      } resolver_t;
/* Old libspf2 API. */  static resolver_t* resolvers;
/* Old libspf2 API. */  static int max_resolvers, first_free_resolver, num_resolvers_inuse;
/* Old libspf2 API. */  static pthread_mutex_t resolver_mutex;
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  init_resolvers( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      max_resolvers = 100;	/* arbitrary */
/* Old libspf2 API. */      resolvers = (resolver_t*) malloc( max_resolvers * sizeof(resolver_t) );
/* Old libspf2 API. */      if( resolvers == (resolver_t*) 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	(void) fprintf( stderr, "%s: out of memory allocating resolver pool\n", argv0 );
/* Old libspf2 API. */  	exit( 1 );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      for( r = 0; r < max_resolvers; ++r )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	resolvers[r].initialized = 0;
/* Old libspf2 API. */  	resolvers[r].inuse = 0;
/* Old libspf2 API. */  	resolvers[r].next_free_resolver = r + 1;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      resolvers[max_resolvers - 1].next_free_resolver = -1;
/* Old libspf2 API. */      first_free_resolver = 0;
/* Old libspf2 API. */      num_resolvers_inuse = 0;
/* Old libspf2 API. */      if( pthread_mutex_init( &resolver_mutex, (pthread_mutexattr_t*) 0 ) != 0 )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	(void) fprintf( stderr, "%s: pthread_mutex_init() failed\n", argv0 );
/* Old libspf2 API. */  	exit( 1 );
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  fini_resolvers( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      for( r = 0; r < max_resolvers; ++r )
/* Old libspf2 API. */  	if( resolvers[r].initialized )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  #ifdef USE_DNS_CACHE
/* Old libspf2 API. */  	    SPF_dns_destroy_config_cache( resolvers[r].spfdcid_c );
/* Old libspf2 API. */  #endif /* USE_DNS_CACHE */
/* Old libspf2 API. */  	    SPF_dns_destroy_config_resolv( resolvers[r].spfdcid_r );
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */      pthread_mutex_destroy( &resolver_mutex );
/* Old libspf2 API. */      free( (void*) resolvers );
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static int
/* Old libspf2 API. */  find_resolver( void )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      int r;
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Find a free slot. */
/* Old libspf2 API. */      pthread_mutex_lock( &resolver_mutex );
/* Old libspf2 API. */      if( num_resolvers_inuse >= max_resolvers )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	/* All are in use, so let's expand the table. */
/* Old libspf2 API. */  	int new_max_resolvers;
/* Old libspf2 API. */  	resolver_t* new_resolvers;
/* Old libspf2 API. */  
/* Old libspf2 API. */  	new_max_resolvers = max_resolvers * 2;
/* Old libspf2 API. */  	syslog( LOG_NOTICE, "enlarging the resolver pool from %d to %d", max_resolvers, new_max_resolvers );
/* Old libspf2 API. */  	new_resolvers = (resolver_t*) realloc( (void*) resolvers, new_max_resolvers * sizeof(resolver_t) );
/* Old libspf2 API. */  	if( new_resolvers == (resolver_t*) 0 )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    pthread_mutex_unlock( &resolver_mutex );
/* Old libspf2 API. */  	    syslog( LOG_ERR, "out of memory enlarging the resolver pool" );
/* Old libspf2 API. */  	    return -1;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	for( r = max_resolvers; r < new_max_resolvers; ++r )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    new_resolvers[r].initialized = 0;
/* Old libspf2 API. */  	    new_resolvers[r].inuse = 0;
/* Old libspf2 API. */  	    new_resolvers[r].next_free_resolver = r + 1;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  	new_resolvers[new_max_resolvers - 1].next_free_resolver = first_free_resolver;
/* Old libspf2 API. */  	first_free_resolver = max_resolvers;
/* Old libspf2 API. */  	resolvers = new_resolvers;
/* Old libspf2 API. */  	max_resolvers = new_max_resolvers;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      if( first_free_resolver == -1 || resolvers[first_free_resolver].inuse )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	pthread_mutex_unlock( &resolver_mutex );
/* Old libspf2 API. */  	syslog( LOG_ERR, "the resolver pool free list is messed up" );
/* Old libspf2 API. */  	return -1;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */      r = first_free_resolver;
/* Old libspf2 API. */      first_free_resolver = resolvers[r].next_free_resolver;
/* Old libspf2 API. */      resolvers[r].next_free_resolver = -1;
/* Old libspf2 API. */      resolvers[r].inuse = 1;
/* Old libspf2 API. */      ++num_resolvers_inuse;
/* Old libspf2 API. */      pthread_mutex_unlock( &resolver_mutex );
/* Old libspf2 API. */  
/* Old libspf2 API. */      /* Is it initialized? */
/* Old libspf2 API. */      if( ! resolvers[r].initialized )
/* Old libspf2 API. */  	{
/* Old libspf2 API. */  	/* Create the DNS objects. */
/* Old libspf2 API. */  	resolvers[r].spfdcid_r = SPF_dns_create_config_resolv( (SPF_dns_config_t) 0, debug );
/* Old libspf2 API. */  	if( resolvers[r].spfdcid_r == (SPF_dns_config_t) 0 )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    syslog( LOG_ERR, "SPF_dns_create_config_resolv() failed" );
/* Old libspf2 API. */  	    free_resolver( r );
/* Old libspf2 API. */  	    return -1;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  #ifdef USE_DNS_CACHE
/* Old libspf2 API. */  	resolvers[r].spfdcid_c = SPF_dns_create_config_cache( resolvers[r].spfdcid_r, 8, debug );
/* Old libspf2 API. */  	if( resolvers[r].spfdcid_c == (SPF_dns_config_t) 0 )
/* Old libspf2 API. */  	    {
/* Old libspf2 API. */  	    syslog( LOG_ERR, "SPF_dns_create_config_cache() failed"  );
/* Old libspf2 API. */  	    free_resolver( r );
/* Old libspf2 API. */  	    return -1;
/* Old libspf2 API. */  	    }
/* Old libspf2 API. */  #endif /* USE_DNS_CACHE */
/* Old libspf2 API. */  	resolvers[r].initialized = 1;
/* Old libspf2 API. */  	}
/* Old libspf2 API. */  
/* Old libspf2 API. */      return r;
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static SPF_dns_config_t
/* Old libspf2 API. */  get_resolver( int r )
/* Old libspf2 API. */      {
/* Old libspf2 API. */  #ifdef USE_DNS_CACHE
/* Old libspf2 API. */      return resolvers[r].spfdcid_c;
/* Old libspf2 API. */  #else /* USE_DNS_CACHE */
/* Old libspf2 API. */      return resolvers[r].spfdcid_r;
/* Old libspf2 API. */  #endif /* USE_DNS_CACHE */
/* Old libspf2 API. */      }
/* Old libspf2 API. */  
/* Old libspf2 API. */  
/* Old libspf2 API. */  static void
/* Old libspf2 API. */  free_resolver( int r )
/* Old libspf2 API. */      {
/* Old libspf2 API. */      pthread_mutex_lock( &resolver_mutex );
/* Old libspf2 API. */      resolvers[r].inuse = 0;
/* Old libspf2 API. */      resolvers[r].next_free_resolver = first_free_resolver;
/* Old libspf2 API. */      first_free_resolver = r;
/* Old libspf2 API. */      --num_resolvers_inuse;
/* Old libspf2 API. */      pthread_mutex_unlock( &resolver_mutex );
/* Old libspf2 API. */      }
/* Old libspf2 API. */  

#else			/* if( SPF_LIB_VERSION_MAJOR == 1 && SPF_LIB_VERSION_MINOR < 2 ) */

/* New libspf2 API. */
/* New libspf2 API. */
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* A static, global pointer to an SPF server object.  We only need one such object no matter	*/
/* New libspf2 API. */  /* how many connections we handle.								*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static SPF_server_t* spf_server;
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Let me get back to you on this one...							*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  struct lib_fallback_s {
/* New libspf2 API. */      /*!!!*/
/* New libspf2 API. */      int junk;
/* New libspf2 API. */      };
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* The per-connection lib_data_s struct.  Contains a couple of pointers and an SPF_result_t.	*/
/* New libspf2 API. */  /* We have a pointer to a structure of this type in the per-connection data structure cd.	*/
/* New libspf2 API. */  /* Memory for the cd structure and contents is malloc()ed by spf_connect() when a connection	*/
/* New libspf2 API. */  /* is made.  We will use cd when testing mail, and hand it sendmail so it can tell us things.	*/
/* New libspf2 API. */  /* The SPF_result_t in this structure is just for debugging at the moment.			*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  struct lib_data_s
/* New libspf2 API. */  {
/* New libspf2 API. */      SPF_request_t*	spf_request;
/* New libspf2 API. */      SPF_response_t*	spf_response;
/* New libspf2 API. */      SPF_result_t	spf_response_result;
/* New libspf2 API. */  };
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_init()								*/
/* New libspf2 API. */  /* Called from:		main()									*/
/* New libspf2 API. */  /* Purpose:		Create the global SPF server object and set the, er, local SPF policy.	*/
/* New libspf2 API. */  /*			The server object is global in this file, which means we don't need a	*/
/* New libspf2 API. */  /*			pointer to it, so we just return 1 if everything works.			*/
/* New libspf2 API. */  /*			If something fails we print a message to stderr and return 0.		*/
/* New libspf2 API. */  /*			This function is called only once per invocation of the milter so we	*/
/* New libspf2 API. */  /*			probably don't need to worry much about memory leaks.			*/
/* New libspf2 API. */  /* Arguments:		None.									*/
/* New libspf2 API. */  /* Returns:		1 on sucess else 0							*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_init( void )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_init()" );
/* New libspf2 API. */      if( recipientmx )
/* New libspf2 API. */      {
/* New libspf2 API. */  	(void) fprintf( stderr, "Sorry, --recipientmx mode is not yet implemented.  GWH, November 2008." );
/* New libspf2 API. */  	return 0;
/* New libspf2 API. */      }
/* New libspf2 API. */      spf_server = SPF_server_new( SPF_DNS_CACHE, debug );
/* New libspf2 API. */      if( spf_server == (SPF_server_t*) 0 )
/* New libspf2 API. */      {
/* New libspf2 API. */  	(void) fprintf( stderr, "%s: SPF_server_new() failed\n", argv0 );
/* New libspf2 API. */  	return 0;
/* New libspf2 API. */      }
/* New libspf2 API. */      if( localpolicy_str != (char*) 0 )
/* New libspf2 API. */      {
/* New libspf2 API. */  	int i;
/* New libspf2 API. */  	SPF_errcode_t err;
/* New libspf2 API. */  	SPF_response_t* lp_response;
/* New libspf2 API. */  	err = SPF_server_set_localpolicy( spf_server, localpolicy_str, trustedforwarders, &lp_response );
/* New libspf2 API. */  	if( err )
/* New libspf2 API. */  	{
/* New libspf2 API. */  	    (void) fprintf( stderr, "%s: error compiling local policy - %d (%s)\n", argv0, err, SPF_strerror( err ) );
/* New libspf2 API. */  	    for( i = 0; i < SPF_response_messages( lp_response ); ++i )
/* New libspf2 API. */  	    {
/* New libspf2 API. */  		SPF_error_t* lp_error = SPF_response_message( lp_response, i );
/* New libspf2 API. */  		(void) fprintf( stderr, "  %s: %s\n",
/* New libspf2 API. */  		    SPF_error_errorp( lp_error ) ? "Error" : "Warning",
/* New libspf2 API. */  		    SPF_error_message( lp_error ) );
/* New libspf2 API. */  	    }
/* New libspf2 API. */  	    SPF_response_free( lp_response );	/* We will exit main() on returning 0 so arguably not necessary. */
/* New libspf2 API. */  	    return 0;
/* New libspf2 API. */  	}
/* New libspf2 API. */  	SPF_response_free( lp_response );
/* New libspf2 API. */      }
/* New libspf2 API. */      if( explanation_str != (char*) 0 )
/* New libspf2 API. */      {
/* New libspf2 API. */  	int err, i;
/* New libspf2 API. */  	SPF_response_t* ex_response;
/* New libspf2 API. */  	err = SPF_server_set_explanation( spf_server, explanation_str, &ex_response );
/* New libspf2 API. */  	if( err )
/* New libspf2 API. */  	{
/* New libspf2 API. */  	    (void) fprintf( stderr, "%s: error compiling explanation - %d (%s)\n", argv0, err, SPF_strerror( err ) );
/* New libspf2 API. */  	    for( i = 0; i < SPF_response_messages( ex_response ); ++i )
/* New libspf2 API. */  	    {
/* New libspf2 API. */  		SPF_error_t* ex_error = SPF_response_message( ex_response, i );
/* New libspf2 API. */  		(void) fprintf( stderr, "  %s: %s\n",
/* New libspf2 API. */  		    SPF_error_errorp( ex_error ) ? "Error" : "Warning",
/* New libspf2 API. */  		    SPF_error_message( ex_error ) );
/* New libspf2 API. */  	    }
/* New libspf2 API. */  	    SPF_response_free( ex_response );	/* We will exit main() on returning 0 so arguably not necessary. */
/* New libspf2 API. */  	    return 0;
/* New libspf2 API. */  	}
/* New libspf2 API. */  	SPF_response_free( ex_response );
/* New libspf2 API. */      }
                            syslog( LOG_DEBUG, "lib_init() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_version()								*/
/* New libspf2 API. */  /* Called from:		main()									*/
/* New libspf2 API. */  /* Purpose:		Return the version of the SPF library.					*/
/* New libspf2 API. */  /* Arguments:		None.									*/
/* New libspf2 API. */  /* Returns:		A pointer to a static character string.					*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static const char*
/* New libspf2 API. */  lib_version( void )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_version()" );
/* New libspf2 API. */      static char version[200];
/* New libspf2 API. */      (void) snprintf( version, sizeof(version), "libspf2-%d.%d.%d", SPF_LIB_VERSION_MAJOR, SPF_LIB_VERSION_MINOR, SPF_LIB_VERSION_PATCH );
/* New libspf2 API. */      return version;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_init_fallback()						FIXME	*/
/* New libspf2 API. */  /* Called from:		add_fallback()								*/
/* New libspf2 API. */  /* Purpose:		Let me get back to you on this one...					*/
/* New libspf2 API. */  /* Arguments:		str: pointer to a character string (which is ignored!)			*/
/* New libspf2 API. */  /* Returns:		Pointer to a lib_fallback_s structure on success else NULL		*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static lib_fallback_t*
/* New libspf2 API. */  lib_init_fallback( const char* str )
/* New libspf2 API. */  {
/* New libspf2 API. */      lib_fallback_t* lf;
                            syslog( LOG_DEBUG, "lib_init_fallback()" );
/* New libspf2 API. */  
/* New libspf2 API. */      lf = (lib_fallback_t*) malloc( sizeof(lib_fallback_t) );
/* New libspf2 API. */      if( lf == (lib_fallback_t*) 0 ) { return (lib_fallback_t*) 0; }
/* New libspf2 API. */      /* FIXME:  Perhaps we should compile the local policy, initialise the structure? (see the old library version) */
                            syslog( LOG_DEBUG, "lib_init_fallback() - done" );
/* New libspf2 API. */      return lf;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_set_local_hostname()						*/
/* New libspf2 API. */  /* Called from:		spf_connect()								*/
/* New libspf2 API. */  /* Purpose:		Initialises the hostname in the libspf2 server data structure		*/
/* New libspf2 API. */  /* Arguments:		ld: pointer to a lib_data_s structure (which is ignored!)		*/
/* New libspf2 API. */  /* Returns:		1 on success else 0							*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_set_local_hostname( lib_data_t* ld )
/* New libspf2 API. */  {   // We use the static file global string pointer local_hostname to initialize spf_server.
                            syslog( LOG_DEBUG, "lib_set_local_hostname()" );
/* New libspf2 API. */	    // The string itself is created by a strdup() call in spf_connect().
/* New libspf2 API. */      // The call to SPF_server_set_rec_dom() actually allocates memory for the hostname in the SPF
/* New libspf2 API. */      // server, but if there's already been some memory allocated for this string it's freed first
/* New libspf2 API. */      // so we don't have to worry about that.  I hope.
/* New libspf2 API. */      // Incidentally it says this (in libspf2) in spf_server.c: "XXX TODO: Free other parts of the structure."
/* New libspf2 API. */      // I just thought I'd mention it.  In case anyone was thinking of doing something about it.  *nudge* *prod* *poke* *dig*
/* New libspf2 API. */      if( SPF_server_set_rec_dom( spf_server, local_hostname ) ) { return 0; }
                            syslog( LOG_DEBUG, "lib_set_local_hostname() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_init_connection_data()						*/
/* New libspf2 API. */  /* Called from:		init_connection_data()							*/
/* New libspf2 API. */  /* Purpose:		Create and initialize a per-connection 'lib_data_s' structure,		*/
/* New libspf2 API. */  /*			and a SPF request object which it will contain.  The lib_data_s		*/
/* New libspf2 API. */  /*			structure also contains a pointer to a SPF response object, we will	*/
/* New libspf2 API. */  /*			only initialize that to a NULL pointer in this function.		*/
/* New libspf2 API. */  /* Arguments:		None.									*/
/* New libspf2 API. */  /* Returns:		Pointer to the lib_data_s structure on success else a NULL poiner.	*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static lib_data_t*
/* New libspf2 API. */  lib_init_connection_data( void )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_init_connection_data()" );
/* New libspf2 API. */      // Create a pointer to a 'lib_data_s' structure, which hopefully will be returned to the caller.
/* New libspf2 API. */      lib_data_t* ld;
/* New libspf2 API. */      /* Create a 'lib_data_s' structure, and point the pointer to it. */
/* New libspf2 API. */      ld = (lib_data_t*) malloc( sizeof(lib_data_t) );
/* New libspf2 API. */      /* Return a NULL pointer if that failed. */
/* New libspf2 API. */      if( ld == (lib_data_t*) 0 ) { return (lib_data_t*) 0; }
/* New libspf2 API. */      /* Create a SPF request object. */
/* New libspf2 API. */      ld->spf_request = SPF_request_new( spf_server );
/* New libspf2 API. */      /* If that failed, free the lib_data_s structure (it's no use to us without a request object), and return a NULL pointer. */
/* New libspf2 API. */      if( ld->spf_request == (SPF_request_t*) 0 ) { free( (void*) ld ); return (lib_data_t*) 0; }
/* New libspf2 API. */      /* Initialize the SPF_response pointer in the lib_data_s structure to NULL. */
/* New libspf2 API. */      ld->spf_response = (SPF_response_t*) 0;
/* New libspf2 API. */      ld->spf_response_result = 0;
                            syslog( LOG_DEBUG, "lib_init_connection_data() - done" );
/* New libspf2 API. */      return ld;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_init_message_data()							*/
/* New libspf2 API. */  /* Called from:		init_message_data(), which is called at each client MAIL FROM command.	*/
/* New libspf2 API. */  /* Purpose:		We don't have to do anything here.					*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure (not used).			*/
/* New libspf2 API. */  /* Returns:		1 (An effort to fool the caller into thinking something succeeded.)	*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_init_message_data( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_init_message_data()" );
/* New libspf2 API. */      /* Nothing to see here, move along please. */
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_set_ipv4()								*/
/* New libspf2 API. */  /* Called from:		spf_connect()								*/
/* New libspf2 API. */  /* Purpose:		Set the IPV4 address of the connecting client in the per-connection	*/
/* New libspf2 API. */  /*			structure (assuming we have an IPV4 connection).			*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			ipv4_addr: Struct representing the connecting IPV4 address (not used).	*/
/* New libspf2 API. */  /*			ipv4_str: String representing the connecting IPV4 address.		*/
/* New libspf2 API. */  /* Returns:		1 on success else 0 on failure.						*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_set_ipv4()" );
/* New libspf2 API. */      if( SPF_request_set_ipv4_str( ld->spf_request, ipv4_str ) ) { return 0; }
                            syslog( LOG_DEBUG, "lib_set_ipv4() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_set_ipv6()								*/
/* New libspf2 API. */  /* Called from:		spf_connect()								*/
/* New libspf2 API. */  /* Purpose:		Set the IPV6 address of the connecting client in the per-connection	*/
/* New libspf2 API. */  /*			structure (assuming we have an IPV6 connection).			*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			ipv6_addr: Struct representing the connecting IPV6 address (not used).	*/
/* New libspf2 API. */  /*			ipv6_str: String representing the connecting IPV6 address.		*/
/* New libspf2 API. */  /* Returns:		1 on success else 0 on failure.						*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_set_ipv6()" );
/* New libspf2 API. */      if( SPF_request_set_ipv6_str( ld->spf_request, ipv6_str ) ) { return 0; }
                            syslog( LOG_DEBUG, "lib_set_ipv6() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_set_helo_hostname()							*/
/* New libspf2 API. */  /* Called from:		spf_helo()								*/
/* New libspf2 API. */  /* Purpose:		Set the client hostname (given by the client in the HELO command) in	*/
/* New libspf2 API. */  /*			the SPF server.  We'll let the SPF library take care of its own memory.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			helo_hostname: String representing the connecting hostname.		*/
/* New libspf2 API. */  /* Returns:		1 on success else 0 on failure.						*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_set_helo_hostname()" );
/* New libspf2 API. */      if( SPF_request_set_helo_dom( ld->spf_request, helo_hostname ) ) { return 0; }
                            syslog( LOG_DEBUG, "lib_set_helo_hostname() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_set_from()								*/
/* New libspf2 API. */  /* Called from:		spf_envfrom()								*/
/* New libspf2 API. */  /* Purpose:		Set the envelope-from address (given by the client in the MAIL FROM	*/
/* New libspf2 API. */  /*			command) in the SPF server.  Let libspf2 take care of its own memory.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			from: String representing the envelope-from address.			*/
/* New libspf2 API. */  /* Returns:		1 on success else 0 on failure.						*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_set_from( lib_data_t* ld, const char* from )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_set_from()" );
/* New libspf2 API. */      if( SPF_request_set_env_from( ld->spf_request, from ) ) { return 0; }
                            syslog( LOG_DEBUG, "lib_set_from() - done" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_do_check_m()						FIXME	*/
/* New libspf2 API. */  /* Called from:		spf_envfrom()								*/
/* New libspf2 API. */  /* Purpose:		Contact the SPF server to check the status of the mail given the	*/
/* New libspf2 API. */  /*			information that we have so far.					*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			from: String representing the envelope-from address, not used here as	*/
/* New libspf2 API. */  /*				the from address has already been set by lib_set_from() above.	*/
/* New libspf2 API. */  /* Returns:		1 on success else 0 on failure.						*/
/* New libspf2 API. */  /* Side effects:	The response_result from the SPF server is written into ld->result but	*/
/* New libspf2 API. */  /*			isn't really needed.							*/
/* New libspf2 API. */  /*			Creates an SPF_response_t object which makes libspf2 allocate memory,	*/
/* New libspf2 API. */  /*			and we need to consider freeing it.  That really sucks.			*/
/* New libspf2 API. */  /* FIXME:		We might want to consider a fallback here!!!				*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_do_check_m( lib_data_t* ld, const char* from )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_do_check_m() - function entry" );
/* New libspf2 API. */      if( ld == (lib_data_t*) 0 ) { syslog( LOG_DEBUG, "lib_do_check_m(): Bad pointer ld" ); return( 0 ); }
/* New libspf2 API. */      ld->spf_response_result = 0;	/* It shouldn't ever be 0. */
/* New libspf2 API. */      /* Check the response object pointer in the per-connection structure hasn't been left untidy. */
/* New libspf2 API. */      if( ld->spf_response != (SPF_response_t*) NULL ) { syslog( LOG_DEBUG, "lib_do_check_m(): Freeing old SPF response object" ); SPF_response_free(ld->spf_response); }
/* New libspf2 API. */      /* Execute the query. */
                            syslog( LOG_DEBUG, "lib_do_check_m(): Sending query to libspf2" );
/* New libspf2 API. */      SPF_errcode_t err = SPF_request_query_mailfrom(ld->spf_request, &ld->spf_response);
/* New libspf2 API. */      if( err )
/* New libspf2 API. */      {
                              syslog( LOG_DEBUG, "lib_do_check_m(): SPF_request_query_mailfrom() returned error code [%d] which might mean [%s] but I'm not sure yet...", err, SPF_strerror(err) );
/* New libspf2 API. */        if( ld->spf_response == (SPF_response_t*) NULL ) { syslog( LOG_DEBUG, "lib_do_check_m(): SPF_request_query_mailfrom() saved NULL pointer in ld->spf_response." ); return( 0 ); }
                              syslog( LOG_DEBUG, "lib_do_check_m(): SPF_request_query_mailfrom() gave ld->spf_response->err=[%d] description is [%s]", ld->spf_response->err, SPF_strerror(ld->spf_response->err) );
			      lib_get_result( ld );
                              syslog( LOG_DEBUG, "lib_do_check_m(): %d errors/warnings returned by libspf", SPF_response_warnings( ld->spf_response ) );
/* New libspf2 API. */      }
                            syslog( LOG_DEBUG, "lib_do_check_m(): Completed libspf2 query" );
/* New libspf2 API. */      /* Grab the result code so we can still see it after we free, as we should, the response object. */
/* New libspf2 API. */      ld->spf_response_result = SPF_response_result(ld->spf_response);
                            syslog( LOG_DEBUG, "lib_do_check_m(): ld->spf_response_result=[%d]", ld->spf_response_result );
/* New libspf2 API. */      /* When we have finished with the response object, we should free it. */
/* New libspf2 API. */      //  SPF_response_free(ld->spf_response);
                            syslog( LOG_DEBUG, "lib_do_check_m() - done - DID NOT FREE RESPONSE OBJECT!" );
/* New libspf2 API. */      return 1;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_do_check_recipient()					FIXME	*/
/* New libspf2 API. */  /* Called from:		spf_envrcpt() if recipientmx=1 (i.e. we're checking the MX records of	*/
/* New libspf2 API. */  /*			recipients).								*/
/* New libspf2 API. */  /* Purpose:		Not yet implemented.  Do what's needed per recipient address for mail	*/
/* New libspf2 API. */  /*			sent to multiple reipients.  Doesn't actually do anything yet.		*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /*			to: String representing the recipient address.				*/
/* New libspf2 API. */  /* Returns:		Zero.  That would normally be considered a failure.			*/
/* New libspf2 API. */  /* FIXME:		We need to implement this function if we want to honour the -r option.	*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_do_check_recipient( lib_data_t* ld, const char* to )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_do_check_recipient()" );
/* New libspf2 API. */      return 0;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_do_check_final()						FIXME	*/
/* New libspf2 API. */  /* Called from:		spf_eoh() i.e. after all the headers have been read, but ONLY IF	*/
/* New libspf2 API. */  /*			recipientmx is TRUE (i.e. we're checking the MX records of recipients).	*/
/* New libspf2 API. */  /* Purpose:		Not yet implemented.  Do what's needed when the end of the mail headers	*/
/* New libspf2 API. */  /*			is reached.  Doesn't actually do anything yet, returns a failure.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /* Returns:		Zero.  That would normally be considered a failure.			*/
/* New libspf2 API. */  /* FIXME:		We need to implement this function if we want to honour the -r option.	*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_do_check_final( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_do_check_final()" );
/* New libspf2 API. */      return 0;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_get_result()							*/
/* New libspf2 API. */  /* Called from:		spf_envfrom(), spf_envrcpt, spf_eoh()					*/
/* New libspf2 API. */  /* Purpose:		Convert the SPF result into something familiar to sendmail.		*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /* Returns:		One of the SPFMILTER_RESULT_XXXX enums, or 0 in case of failure.	*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static int
/* New libspf2 API. */  lib_get_result( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_get_result()" );
/* New libspf2 API. */      if( ld->spf_response == (SPF_response_t*) 0 ) { return 0; }
/* New libspf2 API. */      switch ( SPF_response_result(ld->spf_response) )
/* New libspf2 API. */      {
/* New libspf2 API. */  	case SPF_RESULT_INVALID:	syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_INVALID   " );	return SPFMILTER_RESULT_INVALID;	/* Apparently this shouldn't happen */
/* New libspf2 API. */  	case SPF_RESULT_NONE:		syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_NONE      " );	return SPFMILTER_RESULT_NONE;
/* New libspf2 API. */  	case SPF_RESULT_NEUTRAL:	syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_NEUTRAL   " );	return SPFMILTER_RESULT_NEUTRAL;
/* New libspf2 API. */  	case SPF_RESULT_PASS:		syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_PASS      " );	return SPFMILTER_RESULT_PASS;
/* New libspf2 API. */  	case SPF_RESULT_FAIL:		syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_FAIL      " );	return SPFMILTER_RESULT_FAIL;
/* New libspf2 API. */  	case SPF_RESULT_SOFTFAIL:	syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_SOFTFAIL  " );	return SPFMILTER_RESULT_SOFTFAIL;
/* New libspf2 API. */  	case SPF_RESULT_PERMERROR:	syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_UNKNOWN   " );	return SPFMILTER_RESULT_UNKNOWN;
/* New libspf2 API. */  	case SPF_RESULT_TEMPERROR:	syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_ERROR     " );	return SPFMILTER_RESULT_ERROR;
/* New libspf2 API. */  	default:			syslog( LOG_DEBUG, "lib_get_result(): returning SPFMILTER_RESULT_AARRGGHH  " );	return 0;	/* Shouldn't be reached. */
/* New libspf2 API. */      }
/* New libspf2 API. */      /* Never reached. */
/* New libspf2 API. */      return 0;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_get_explanation()						FIXME	*/
/* New libspf2 API. */  /* Called from:		handle_result() if SPFMILTER_ACTION_REJECT				*/
/* New libspf2 API. */  /* Purpose:		Look up a character string to explain the rejection of the mail.	*/
/* New libspf2 API. */  /*			The string returned by this function will be returned to the client.	*/
/* New libspf2 API. */  /*			Doesn't actually do anything yet, and it doesn't matter as a general-	*/
/* New libspf2 API. */  /*			purpose string will be used by the calling routine in that case.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /* Returns:		NULL pointer.  It's not a big deal.					*/
/* New libspf2 API. */  /* FIXME:		Eventually.								*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static const char*
/* New libspf2 API. */  lib_get_explanation( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_get_explanation()" );
/* New libspf2 API. */      return (char*) 0;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_get_error()							FIXME	*/
/* New libspf2 API. */  /* Called from:		handle_result() if SPFMILTER_ACTION_TEMPFAIL				*/
/* New libspf2 API. */  /* Purpose:		Look up a character string to explain the tempfailing of the mail.	*/
/* New libspf2 API. */  /*			Doesn't actually do anything yet, and it doesn't matter as the message	*/
/* New libspf2 API. */  /*			containing the returned string would only be used in a syslog call.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /* Returns:		NULL pointer.  It's not a big deal.					*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static const char*
/* New libspf2 API. */  lib_get_error( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_get_error()" );
/* New libspf2 API. */      return (char*) 0;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_fini_message_data()							*/
/* New libspf2 API. */  /* Called from:		fini_message_data()							*/
/* New libspf2 API. */  /* Purpose:		We'd clean up, for example call free() for some data structures,	*/
/* New libspf2 API. */  /*			but we don't actually do anything as no action is necessary.		*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure (ignored).			*/
/* New libspf2 API. */  /* Returns:		Nothing whatsoever.  Zilch.  Nada.  Rien.				*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static void
/* New libspf2 API. */  lib_fini_message_data( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_fini_message_data()" );
/* New libspf2 API. */    ;
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_fini_connection_data()						*/
/* New libspf2 API. */  /* Called from:		fini_connection_data()							*/
/* New libspf2 API. */  /* Purpose:		Clean up after ourselves.  Free a lib_data_s structure and (er, well,	*/
/* New libspf2 API. */  /*			first, obviously) memory allocated for things its pointers point to.	*/
/* New libspf2 API. */  /* Arguments:		ld: Pointer to the per-connection structure.				*/
/* New libspf2 API. */  /* Returns:		Nothing.  There's nothing we can do if it fails anyway.			*/
/* New libspf2 API. */  /* WARNING:		The connection data structures are malloc()ed for each connection	*/
/* New libspf2 API. */  /*			during the life of the milter daemon.  This routine must free the data	*/
/* New libspf2 API. */  /*			properly after each connection, or the milter will leak memory.		*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static void
/* New libspf2 API. */  lib_fini_connection_data( lib_data_t* ld )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_fini_connection_data()" );
/* New libspf2 API. */      if( ld != (lib_data_t*) 0 )
/* New libspf2 API. */      {
/* New libspf2 API. */  	if( ld->spf_request != (SPF_request_t*) 0 ) { SPF_request_free( ld->spf_request ); }
/* New libspf2 API. */  	if( ld->spf_response != (SPF_response_t*) 0 ) { SPF_response_free( ld->spf_response ); }
/* New libspf2 API. */  	free( (void*) ld );
/* New libspf2 API. */      }
                            syslog( LOG_DEBUG, "lib_fini_connection_data() - done" );
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_fini_fallback()							*/
/* New libspf2 API. */  /* Called from:		fini_fallback()								*/
/* New libspf2 API. */  /* Purpose:		Clean up after ourselves.  Free a lib_fallback_s structure.		*/
/* New libspf2 API. */  /* Arguments:		lf: Pointer to the structure we're going to blow away.  Often nested.	*/
/* New libspf2 API. */  /* Returns:		Nothing.  There's nothing we can do if it fails anyway.			*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static void
/* New libspf2 API. */  lib_fini_fallback( lib_fallback_t* lf )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_fini_fallback()" );
/* New libspf2 API. */      if( lf != (lib_fallback_t*) 0 )
/* New libspf2 API. */      {
/* New libspf2 API. */  	/* FIXME */		// Is there anything you want to tell me before we begin?
/* New libspf2 API. */  	free( (void*) lf );	// If the structure IS nested, this will leak like a sieve,
/* New libspf2 API. */      }				// but as this is only called at the end of main() there's
                            syslog( LOG_DEBUG, "lib_fini_fallback() - done" );
/* New libspf2 API. */  }				// no need to worry, the OS will do it for us anyway.
/* New libspf2 API. */  
/* New libspf2 API. */  
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  /* Function name:	lib_fini()								*/
/* New libspf2 API. */  /* Called from:		main()									*/
/* New libspf2 API. */  /* Purpose:		Free the SPF server - but the OS will do this for us if we don't...	*/
/* New libspf2 API. */  /* Arguments:		None.									*/
/* New libspf2 API. */  /* Returns:		Nothing.  It's all over now.						*/
/* New libspf2 API. */  /************************************************************************************************/
/* New libspf2 API. */  static void
/* New libspf2 API. */  lib_fini( void )
/* New libspf2 API. */  {
                            syslog( LOG_DEBUG, "lib_fini() - freeing SPF server" );
/* New libspf2 API. */      SPF_server_free( spf_server );
                            syslog( LOG_DEBUG, "lib_fini() - done" );
/* New libspf2 API. */  }
/* New libspf2 API. */  
/* New libspf2 API. */  #endif	/* if( SPF_LIB_VERSION_MAJOR == 1 && SPF_LIB_VERSION_MINOR < 2 ) */
/* New libspf2 API. */  
/* New libspf2 API. */  #endif /* USE_LIBSPF2 */
/* New libspf2 API. */  

// EOF: spfmilter.c
